# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""This module defines an abstract problem class."""

import unified_planning as up
from typing import Optional


class AbstractProblem:
    """This is an abstract class that represents a generic planning problem"""

    def __init__(self, name: str = None, env: "up.environment.Environment" = None):
        self._env = up.environment.get_env(env)
        self._name = name

    @property
    def env(self) -> "up.environment.Environment":
        """Returns the problem environment."""
        return self._env

    @property
    def name(self) -> Optional[str]:
        """Returns the problem name."""
        return self._name

    @name.setter
    def name(self, new_name: str):
        """Sets the problem name."""
        self._name = new_name

    @property
    def kind(self) -> "up.model.problem_kind.ProblemKind":
        """Returns the ProblemKind of this problem."""
        raise NotImplementedError

    def has_name(self, name: str) -> bool:
        """Returns True iff the given name is already used inside this problem."""
        raise NotImplementedError

    def normalize_plan(self, plan: "up.plans.Plan") -> "up.plans.Plan":
        """Normalizes the given plan, that is potentially the result of another
        problem, updating the object references present in it with the ones of
        this problem which are syntactically equal."""
        raise NotImplementedError



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""This module defines the engine interface."""

from unified_planning.model import ProblemKind
from unified_planning.engines.credits import Credits
from typing import Optional


class EngineMeta(type):
    def __new__(cls, name, bases, dct):
        obj = type.__new__(cls, name, bases, dct)
        for om in ["oneshot_planner", "plan_validator", "compiler", "simulator"]:
            if (
                not hasattr(obj, "is_" + om)
                and name != "Engine"
                and name != "MetaEngine"
            ):
                setattr(obj, "is_" + om, lambda: False)
        return obj


class Engine(metaclass=EngineMeta):
    """Represents the engine interface."""

    @property
    def name(self) -> str:
        raise NotImplementedError

    @staticmethod
    def supported_kind() -> ProblemKind:
        raise NotImplementedError

    @staticmethod
    def supports(problem_kind: "ProblemKind") -> bool:
        raise NotImplementedError

    @staticmethod
    def get_credits(**kwargs) -> Optional[Credits]:
        """
        This method returns the credits for this engine, that will be printed when the engine is used.
        If this function returns None, it means no credits to print.

        The **kwargs parameters are the same used in this engine to communicate
         the specific options for this Engine instance.
        """
        return None

    def destroy(self):
        pass

    def __enter__(self):
        """Manages entering a Context (i.e., with statement)"""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Manages exiting from Context (i.e., with statement)"""
        self.destroy()



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import unified_planning as up
from unified_planning.environment import Environment, get_env
from typing import Callable, Optional, Tuple
from enum import Enum, auto


"""This module defines the different plan classes."""


class ActionInstance:
    """Represents an action instance with the actual parameters.

    NOTE: two action instances of the same action with the same parameters are
    considered different as it is possible to have the same action twice in a plan."""

    def __init__(
        self, action: "up.model.Action", params: Tuple["up.model.FNode", ...] = tuple()
    ):
        assert len(action.parameters) == len(params)
        self._action = action
        self._params = tuple(params)

    def __repr__(self) -> str:
        s = []
        if len(self._params) > 0:
            s.append("(")
            first = True
            for p in self._params:
                if not first:
                    s.append(", ")
                s.append(str(p))
                first = False
            s.append(")")
        return self._action.name + "".join(s)

    @property
    def action(self) -> "up.model.Action":
        """Returns the action."""
        return self._action

    @property
    def actual_parameters(self) -> Tuple["up.model.FNode", ...]:
        """Returns the actual parameters."""
        return self._params

    def is_semantically_equivalent(self, oth: "ActionInstance") -> bool:
        """This method returns True Iff the 2 Action Instances have the same semantic.

        NOTE: This is different from __eq__; there the 2 Action Instances need to be exactly the same object."""
        return self.action == oth.action and self._params == oth._params


class PlanKind(Enum):
    SEQUENTIAL_PLAN = auto()
    TIME_TRIGGERED_PLAN = auto()
    PARTIAL_ORDER_PLAN = auto()


class Plan:
    """Represents a generic plan."""

    def __init__(
        self, kind: PlanKind, environment: Optional["Environment"] = None
    ) -> None:
        self._kind = kind
        self._environment = get_env(environment)

    @property
    def environment(self) -> "Environment":
        """Return this plan environment."""
        return self._environment

    @property
    def kind(self) -> PlanKind:
        """Returns the plan kind"""
        return self._kind

    def replace_action_instances(
        self, replace_function: Callable[[ActionInstance], ActionInstance]
    ) -> "Plan":
        """This function takes a function from ActionInstance to ActionInstance and returns a new Plan
        that have the ActionInstance modified by the "replace_function" function."""
        raise NotImplementedError



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""This module defines the problem class."""


import unified_planning as up
from unified_planning.model.abstract_problem import AbstractProblem
from unified_planning.model.mixins import (
    ActionsSetMixin,
    FluentsSetMixin,
    ObjectsSetMixin,
    UserTypesSetMixin,
)
from unified_planning.model.expression import ConstantExpression
from unified_planning.model.operators import OperatorKind
from unified_planning.model.types import domain_size, domain_item
from unified_planning.exceptions import (
    UPProblemDefinitionError,
    UPTypeError,
    UPExpressionDefinitionError,
)
from fractions import Fraction
from typing import List, Dict, Set, Union, cast


class Problem(
    AbstractProblem,
    UserTypesSetMixin,
    FluentsSetMixin,
    ActionsSetMixin,
    ObjectsSetMixin,
):
    """Represents a, action based planning problem."""

    def __init__(
        self,
        name: str = None,
        env: "up.environment.Environment" = None,
        *,
        initial_defaults: Dict["up.model.types.Type", "ConstantExpression"] = {},
    ):
        AbstractProblem.__init__(self, name, env)
        UserTypesSetMixin.__init__(self, self.has_name)
        FluentsSetMixin.__init__(
            self, self.env, self._add_user_type, self.has_name, initial_defaults
        )
        ActionsSetMixin.__init__(self, self.env, self._add_user_type, self.has_name)
        ObjectsSetMixin.__init__(self, self.env, self._add_user_type, self.has_name)
        self._operators_extractor = up.model.walkers.OperatorsExtractor()
        self._initial_value: Dict["up.model.fnode.FNode", "up.model.fnode.FNode"] = {}
        self._timed_effects: Dict[
            "up.model.timing.Timing", List["up.model.effect.Effect"]
        ] = {}
        self._timed_goals: Dict[
            "up.model.timing.TimeInterval", List["up.model.fnode.FNode"]
        ] = {}
        self._goals: List["up.model.fnode.FNode"] = list()
        self._metrics: List["up.model.metrics.PlanQualityMetric"] = []

    def __repr__(self) -> str:
        s = []
        if not self.name is None:
            s.append(f"problem name = {str(self.name)}\n\n")
        if len(self.user_types) > 0:
            s.append(f"types = {str(list(self.user_types))}\n\n")
        s.append("fluents = [\n")
        for f in self.fluents:
            s.append(f"  {str(f)}\n")
        s.append("]\n\n")
        s.append("actions = [\n")
        for a in self.actions:
            s.append(f"  {str(a)}\n")
        s.append("]\n\n")
        if len(self.user_types) > 0:
            s.append("objects = [\n")
            for ty in self.user_types:
                s.append(f"  {str(ty)}: {str(list(self.objects(ty)))}\n")
            s.append("]\n\n")
        s.append("initial fluents default = [\n")
        for f in self._fluents:
            if f in self._fluents_defaults:
                v = self._fluents_defaults[f]
                s.append(f"  {str(f)} := {str(v)}\n")
        s.append("]\n\n")
        s.append("initial values = [\n")
        for k, v in self._initial_value.items():
            s.append(f"  {str(k)} := {str(v)}\n")
        s.append("]\n\n")
        if len(self.timed_effects) > 0:
            s.append("timed effects = [\n")
            for t, el in self.timed_effects.items():
                s.append(f"  {str(t)} :\n")
                for e in el:
                    s.append(f"    {str(e)}\n")
            s.append("]\n\n")
        if len(self.timed_goals) > 0:
            s.append("timed goals = [\n")
            for i, gl in self.timed_goals.items():
                s.append(f"  {str(i)} :\n")
                for g in gl:
                    s.append(f"    {str(g)}\n")
            s.append("]\n\n")
        s.append("goals = [\n")
        for g in self.goals:
            s.append(f"  {str(g)}\n")
        s.append("]\n\n")
        if len(self.quality_metrics) > 0:
            s.append("quality metrics = [\n")
            for qm in self.quality_metrics:
                s.append(f"  {str(qm)}\n")
            s.append("]\n\n")
        return "".join(s)

    def __eq__(self, oth: object) -> bool:
        if not (isinstance(oth, Problem)) or self._env != oth._env:
            return False
        if self.kind != oth.kind or self._name != oth._name:
            return False
        if set(self._fluents) != set(oth._fluents) or set(self._goals) != set(
            oth._goals
        ):
            return False
        if set(self._user_types) != set(oth._user_types) or set(self._objects) != set(
            oth._objects
        ):
            return False
        if set(self._actions) != set(oth._actions):
            return False
        oth_initial_values = oth.initial_values
        if len(self.initial_values) != len(oth_initial_values):
            return False
        for fluent, value in self.initial_values.items():
            oth_value = oth_initial_values.get(fluent, None)
            if oth_value is None:
                return False
            elif value != oth_value:
                return False
        if len(self._timed_effects) != len(oth._timed_effects):
            return False
        for t, tel in self._timed_effects.items():
            oth_tel = oth._timed_effects.get(t, None)
            if oth_tel is None:
                return False
            elif set(tel) != set(oth_tel):
                return False
        if len(self._timed_goals) != len(oth._timed_goals):
            return False
        for i, tgl in self._timed_goals.items():
            oth_tgl = oth._timed_goals.get(i, None)
            if oth_tgl is None:
                return False
            elif set(tgl) != set(oth_tgl):
                return False
        return True

    def __hash__(self) -> int:
        res = hash(self._kind) + hash(self._name)
        for f in self._fluents:
            res += hash(f)
        for a in self._actions:
            res += hash(a)
        for ut in self._user_types:
            res += hash(ut)
        for o in self._objects:
            res += hash(o)
        for iv in self.initial_values.items():
            res += hash(iv)
        for t, el in self._timed_effects.items():
            res += hash(t)
            for e in set(el):
                res += hash(e)
        for i, gl in self._timed_goals.items():
            res += hash(i)
            for g in set(gl):
                res += hash(g)
        for g in self._goals:
            res += hash(g)
        return res

    def clone(self):
        new_p = Problem(self._name, self._env)
        new_p._fluents = self._fluents[:]
        new_p._actions = [a.clone() for a in self._actions]
        new_p._user_types = self._user_types[:]
        new_p._user_types_hierarchy = self._user_types_hierarchy.copy()
        new_p._objects = self._objects[:]
        new_p._initial_value = self._initial_value.copy()
        new_p._timed_effects = {
            t: [e.clone() for e in el] for t, el in self._timed_effects.items()
        }
        new_p._timed_goals = {i: [g for g in gl] for i, gl in self._timed_goals.items()}
        new_p._goals = self._goals[:]
        new_p._metrics = []
        for m in self._metrics:
            if isinstance(m, up.model.metrics.MinimizeActionCosts):
                costs = {new_p.action(a.name): c for a, c in m.costs.items()}
                new_p._metrics.append(up.model.metrics.MinimizeActionCosts(costs))
            else:
                new_p._metrics.append(m)
        new_p._initial_defaults = self._initial_defaults.copy()
        new_p._fluents_defaults = self._fluents_defaults.copy()
        return new_p

    def has_name(self, name: str) -> bool:
        """Returns true if the name is in the problem."""
        return (
            self.has_action(name)
            or self.has_fluent(name)
            or self.has_object(name)
            or self.has_type(name)
        )

    def normalize_plan(self, plan: "up.plans.Plan") -> "up.plans.Plan":
        """Normalizes the given plan updating the action and object instances."""
        return plan.replace_action_instances(self._replace_action_instance)

    def _replace_action_instance(
        self, action_instance: "up.plans.ActionInstance"
    ) -> "up.plans.ActionInstance":
        em = self.env.expression_manager
        new_a = self.action(action_instance.action.name)
        params = []
        for p in action_instance.actual_parameters:
            if p.is_object_exp():
                obj = self.object(p.object().name)
                params.append(em.ObjectExp(obj))
            elif p.is_bool_constant():
                params.append(em.Bool(p.is_true()))
            elif p.is_int_constant():
                params.append(em.Int(cast(int, p.constant_value())))
            elif p.is_real_constant():
                params.append(em.Real(cast(Fraction, p.constant_value())))
            else:
                raise NotImplementedError
        return up.plans.ActionInstance(new_a, tuple(params))

    def get_static_fluents(self) -> Set["up.model.fluent.Fluent"]:
        """Returns the set of the static fluents.

        Static fluents are those who can't change their values because they never
        appear in the "fluent" field of an effect, therefore there are no Actions
        in the Problem that can change their value."""
        static_fluents: Set["up.model.fluent.Fluent"] = set(self._fluents)
        for a in self._actions:
            if isinstance(a, up.model.action.InstantaneousAction):
                for e in a.effects:
                    static_fluents.discard(e.fluent.fluent())
                if a.simulated_effect is not None:
                    for f in a.simulated_effect.fluents:
                        static_fluents.discard(f.fluent())
            elif isinstance(a, up.model.action.DurativeAction):
                for el in a.effects.values():
                    for e in el:
                        static_fluents.discard(e.fluent.fluent())
                for _, se in a.simulated_effects.items():
                    for f in se.fluents:
                        static_fluents.discard(f.fluent())
            else:
                raise NotImplementedError
        for el in self._timed_effects.values():
            for e in el:
                if e.fluent.fluent() in static_fluents:
                    static_fluents.remove(e.fluent.fluent())
        return static_fluents

    def set_initial_value(
        self,
        fluent: Union["up.model.fnode.FNode", "up.model.fluent.Fluent"],
        value: Union[
            "up.model.fnode.FNode",
            "up.model.fluent.Fluent",
            "up.model.object.Object",
            bool,
            int,
            float,
            Fraction,
        ],
    ):
        """Sets the initial value for the given fluent."""
        fluent_exp, value_exp = self._env.expression_manager.auto_promote(fluent, value)
        if not fluent_exp.type.is_compatible(value_exp.type):
            raise UPTypeError("Initial value assignment has not compatible types!")
        self._initial_value[fluent_exp] = value_exp

    def initial_value(
        self, fluent: Union["up.model.fnode.FNode", "up.model.fluent.Fluent"]
    ) -> "up.model.fnode.FNode":
        """Gets the initial value of the given fluent."""
        (fluent_exp,) = self._env.expression_manager.auto_promote(fluent)
        for a in fluent_exp.args:
            if not a.is_constant():
                raise UPExpressionDefinitionError(
                    f"Impossible to return the initial value of a fluent expression with no constant arguments: {fluent_exp}."
                )
        if fluent_exp in self._initial_value:
            return self._initial_value[fluent_exp]
        elif fluent_exp.fluent() in self._fluents_defaults:
            return self._fluents_defaults[fluent_exp.fluent()]
        else:
            raise UPProblemDefinitionError("Initial value not set!")

    def _get_ith_fluent_exp(
        self, fluent: "up.model.fluent.Fluent", domain_sizes: List[int], idx: int
    ) -> "up.model.fnode.FNode":
        """Returns the ith ground fluent expression."""
        quot = idx
        rem = 0
        actual_parameters = []
        for i, p in enumerate(fluent.signature):
            ds = domain_sizes[i]
            rem = quot % ds
            quot //= ds
            v = domain_item(self, p.type, rem)
            actual_parameters.append(v)
        return fluent(*actual_parameters)

    @property
    def initial_values(self) -> Dict["up.model.fnode.FNode", "up.model.fnode.FNode"]:
        """Gets the initial value of the fluents.

        IMPORTANT NOTE: this property does a lot of computation, so it should be called as
        seldom as possible."""
        res = self._initial_value
        for f in self._fluents:
            if f.arity == 0:
                f_exp = self._env.expression_manager.FluentExp(f)
                res[f_exp] = self.initial_value(f_exp)
            else:
                ground_size = 1
                domain_sizes = []
                for p in f.signature:
                    ds = domain_size(self, p.type)
                    domain_sizes.append(ds)
                    ground_size *= ds
                for i in range(ground_size):
                    f_exp = self._get_ith_fluent_exp(f, domain_sizes, i)
                    res[f_exp] = self.initial_value(f_exp)
        return res

    @property
    def explicit_initial_values(
        self,
    ) -> Dict["up.model.fnode.FNode", "up.model.fnode.FNode"]:
        """Returns the problem's defined initial values.
        IMPORTANT NOTE: For all the initial values of hte problem use Problem.initial_values."""
        return self._initial_value

    def add_timed_goal(
        self,
        interval: Union["up.model.timing.Timing", "up.model.timing.TimeInterval"],
        goal: Union["up.model.fnode.FNode", "up.model.fluent.Fluent", bool],
    ):
        """Adds a timed goal."""
        assert (
            isinstance(goal, bool) or goal.environment == self._env
        ), "timed_goal does not have the same environment of the problem"
        if isinstance(interval, up.model.Timing):
            interval = up.model.TimePointInterval(interval)
        if (interval.lower.is_from_end() and interval.lower.delay > 0) or (
            interval.upper.is_from_end() and interval.upper.delay > 0
        ):
            raise UPProblemDefinitionError(
                "Problem timing can not be `end - k` with k > 0."
            )
        (goal_exp,) = self._env.expression_manager.auto_promote(goal)
        assert self._env.type_checker.get_type(goal_exp).is_bool_type()
        if interval in self._timed_goals:
            if goal_exp not in self._timed_goals[interval]:
                self._timed_goals[interval].append(goal_exp)
        else:
            self._timed_goals[interval] = [goal_exp]

    @property
    def timed_goals(
        self,
    ) -> Dict["up.model.timing.TimeInterval", List["up.model.fnode.FNode"]]:
        """Returns the timed goals."""
        return self._timed_goals

    def clear_timed_goals(self):
        """Removes the timed goals."""
        self._timed_goals = {}

    def add_timed_effect(
        self,
        timing: "up.model.timing.Timing",
        fluent: Union["up.model.fnode.FNode", "up.model.fluent.Fluent"],
        value: "up.model.expression.Expression",
        condition: "up.model.expression.BoolExpression" = True,
    ):
        """Adds the given timed effect."""
        if timing.is_from_end():
            raise UPProblemDefinitionError(
                f"Timing used in timed effect cannot be EndTiming."
            )
        (
            fluent_exp,
            value_exp,
            condition_exp,
        ) = self._env.expression_manager.auto_promote(fluent, value, condition)
        assert fluent_exp.is_fluent_exp()
        if not self._env.type_checker.get_type(condition_exp).is_bool_type():
            raise UPTypeError("Effect condition is not a Boolean condition!")
        if not fluent_exp.type.is_compatible(value_exp.type):
            raise UPTypeError("Timed effect has not compatible types!")
        self._add_effect_instance(
            timing, up.model.effect.Effect(fluent_exp, value_exp, condition_exp)
        )

    def add_increase_effect(
        self,
        timing: "up.model.timing.Timing",
        fluent: Union["up.model.fnode.FNode", "up.model.fluent.Fluent"],
        value: "up.model.expression.Expression",
        condition: "up.model.expression.BoolExpression" = True,
    ):
        """Adds the given timed increase effect."""
        (
            fluent_exp,
            value_exp,
            condition_exp,
        ) = self._env.expression_manager.auto_promote(fluent, value, condition)
        assert fluent_exp.is_fluent_exp()
        if not self._env.type_checker.get_type(condition_exp).is_bool_type():
            raise UPTypeError("Effect condition is not a Boolean condition!")
        if not fluent_exp.type.is_compatible(value_exp.type):
            raise UPTypeError("Timed effect has not compatible types!")
        self._add_effect_instance(
            timing,
            up.model.effect.Effect(
                fluent_exp,
                value_exp,
                condition_exp,
                kind=up.model.effect.EffectKind.INCREASE,
            ),
        )

    def add_decrease_effect(
        self,
        timing: "up.model.timing.Timing",
        fluent: Union["up.model.fnode.FNode", "up.model.fluent.Fluent"],
        value: "up.model.expression.Expression",
        condition: "up.model.expression.BoolExpression" = True,
    ):
        """Adds the given timed decrease effect."""
        (
            fluent_exp,
            value_exp,
            condition_exp,
        ) = self._env.expression_manager.auto_promote(fluent, value, condition)
        assert fluent_exp.is_fluent_exp()
        if not self._env.type_checker.get_type(condition_exp).is_bool_type():
            raise UPTypeError("Effect condition is not a Boolean condition!")
        if not fluent_exp.type.is_compatible(value_exp.type):
            raise UPTypeError("Timed effect has not compatible types!")
        self._add_effect_instance(
            timing,
            up.model.effect.Effect(
                fluent_exp,
                value_exp,
                condition_exp,
                kind=up.model.effect.EffectKind.DECREASE,
            ),
        )

    def _add_effect_instance(
        self, timing: "up.model.timing.Timing", effect: "up.model.effect.Effect"
    ):
        assert (
            effect.environment == self._env
        ), "effect does not have the same environment of the problem"
        if timing in self._timed_effects:
            if effect not in self._timed_effects[timing]:
                self._timed_effects[timing].append(effect)
        else:
            self._timed_effects[timing] = [effect]

    @property
    def timed_effects(
        self,
    ) -> Dict["up.model.timing.Timing", List["up.model.effect.Effect"]]:
        """Returns the timed effects."""
        return self._timed_effects

    def clear_timed_effects(self):
        """Removes the timed effects."""
        self._timed_effects = {}

    def add_goal(
        self, goal: Union["up.model.fnode.FNode", "up.model.fluent.Fluent", bool]
    ):
        """Adds a goal."""
        assert (
            isinstance(goal, bool) or goal.environment == self._env
        ), "goal does not have the same environment of the problem"
        (goal_exp,) = self._env.expression_manager.auto_promote(goal)
        assert self._env.type_checker.get_type(goal_exp).is_bool_type()
        if goal_exp != self._env.expression_manager.TRUE():
            self._goals.append(goal_exp)

    @property
    def goals(self) -> List["up.model.fnode.FNode"]:
        """Returns the goals."""
        return self._goals

    def clear_goals(self):
        """Removes the goals."""
        self._goals = []

    def add_quality_metric(self, metric: "up.model.metrics.PlanQualityMetric"):
        """Adds a quality metric"""
        self._metrics.append(metric)

    @property
    def quality_metrics(self) -> List["up.model.metrics.PlanQualityMetric"]:
        """Returns the quality metrics"""
        return self._metrics

    def clear_quality_metrics(self):
        """Removes the quality metrics"""
        self._metrics = []

    @property
    def kind(self) -> "up.model.problem_kind.ProblemKind":
        """Returns the problem kind of this planning problem.

        IMPORTANT NOTE: this property does a lot of computation, so it should be called as
        seldom as possible."""
        self._kind = up.model.problem_kind.ProblemKind()
        self._kind.set_problem_class("ACTION_BASED")
        for fluent in self._fluents:
            self._update_problem_kind_fluent(fluent)
        for action in self._actions:
            self._update_problem_kind_action(action)
        if len(self._timed_effects) > 0:
            self._kind.set_time("CONTINUOUS_TIME")
            self._kind.set_time("TIMED_EFFECT")
        for effect_list in self._timed_effects.values():
            for effect in effect_list:
                self._update_problem_kind_effect(effect)
        if len(self._timed_goals) > 0:
            self._kind.set_time("TIMED_GOALS")
            self._kind.set_time("CONTINUOUS_TIME")
        for goal_list in self._timed_goals.values():
            for goal in goal_list:
                self._update_problem_kind_condition(goal)
        for goal in self._goals:
            self._update_problem_kind_condition(goal)
        for metric in self._metrics:
            if isinstance(
                metric, up.model.metrics.MinimizeExpressionOnFinalState
            ) or isinstance(metric, up.model.metrics.MaximizeExpressionOnFinalState):
                self._kind.set_quality_metrics("FINAL_VALUE")
            elif isinstance(metric, up.model.metrics.MinimizeActionCosts):
                self._kind.set_quality_metrics("ACTIONS_COST")
            elif isinstance(metric, up.model.metrics.MinimizeMakespan):
                self._kind.set_quality_metrics("MAKESPAN")
            elif isinstance(metric, up.model.metrics.MinimizeSequentialPlanLength):
                self._kind.set_quality_metrics("PLAN_LENGTH")
            elif isinstance(metric, up.model.metrics.Oversubscription):
                self._kind.set_quality_metrics("OVERSUBSCRIPTION")
            else:
                assert False, "Unknown quality metric"
        return self._kind

    def _update_problem_kind_effect(self, e: "up.model.effect.Effect"):
        if e.is_conditional():
            self._update_problem_kind_condition(e.condition)
            self._kind.set_effects_kind("CONDITIONAL_EFFECTS")
        if e.is_increase():
            self._kind.set_effects_kind("INCREASE_EFFECTS")
        elif e.is_decrease():
            self._kind.set_effects_kind("DECREASE_EFFECTS")

    def _update_problem_kind_condition(self, exp: "up.model.fnode.FNode"):
        ops = self._operators_extractor.get(exp)
        if OperatorKind.EQUALS in ops:
            self._kind.set_conditions_kind("EQUALITY")
        if OperatorKind.NOT in ops:
            self._kind.set_conditions_kind("NEGATIVE_CONDITIONS")
        if OperatorKind.OR in ops:
            self._kind.set_conditions_kind("DISJUNCTIVE_CONDITIONS")
        if OperatorKind.EXISTS in ops:
            self._kind.set_conditions_kind("EXISTENTIAL_CONDITIONS")
        if OperatorKind.FORALL in ops:
            self._kind.set_conditions_kind("UNIVERSAL_CONDITIONS")

    def _update_problem_kind_type(self, type: "up.model.types.Type"):
        if type.is_user_type():
            self._kind.set_typing("FLAT_TYPING")
            if cast(up.model.types._UserType, type).father is not None:
                self._kind.set_typing("HIERARCHICAL_TYPING")
        elif type.is_int_type():
            self._kind.set_numbers("DISCRETE_NUMBERS")
        elif type.is_real_type():
            self._kind.set_numbers("CONTINUOUS_NUMBERS")

    def _update_problem_kind_fluent(self, fluent: "up.model.fluent.Fluent"):
        self._update_problem_kind_type(fluent.type)
        if fluent.type.is_int_type() or fluent.type.is_real_type():
            self._kind.set_fluents_type("NUMERIC_FLUENTS")
        elif fluent.type.is_user_type():
            self._kind.set_fluents_type("OBJECT_FLUENTS")
        for p in fluent.signature:
            self._update_problem_kind_type(p.type)

    def _update_problem_kind_action(self, action: "up.model.action.Action"):
        for p in action.parameters:
            self._update_problem_kind_type(p.type)
        if isinstance(action, up.model.action.InstantaneousAction):
            for c in action.preconditions:
                self._update_problem_kind_condition(c)
            for e in action.effects:
                self._update_problem_kind_effect(e)
            if action.simulated_effect is not None:
                self._kind.set_simulated_entities("SIMULATED_EFFECTS")
        elif isinstance(action, up.model.action.DurativeAction):
            lower, upper = action.duration.lower, action.duration.upper
            if lower != upper:
                self._kind.set_time("DURATION_INEQUALITIES")
            free_vars = self.env.free_vars_extractor.get(
                lower
            ) | self.env.free_vars_extractor.get(upper)
            if len(free_vars) > 0:
                static_fluents = self.get_static_fluents()
                only_static = True
                for fv in free_vars:
                    if fv.fluent() not in static_fluents:
                        only_static = False
                        break
                if only_static:
                    self._kind.set_expression_duration("STATIC_FLUENTS_IN_DURATION")
                else:
                    self._kind.set_expression_duration("FLUENTS_IN_DURATION")
            for i, lc in action.conditions.items():
                if i.lower.delay != 0 or i.upper.delay != 0:
                    self._kind.set_time("INTERMEDIATE_CONDITIONS_AND_EFFECTS")
                for c in lc:
                    self._update_problem_kind_condition(c)
            for t, le in action.effects.items():
                if t.delay != 0:
                    self._kind.set_time("INTERMEDIATE_CONDITIONS_AND_EFFECTS")
                for e in le:
                    self._update_problem_kind_effect(e)
            if len(action.simulated_effects) > 0:
                self._kind.set_simulated_entities("SIMULATED_EFFECTS")
            self._kind.set_time("CONTINUOUS_TIME")
        else:
            raise NotImplementedError



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import unified_planning.environment
import unified_planning.model.walkers as walkers
from unified_planning.exceptions import UPUnreachableCodeError
from unified_planning.model.fnode import FNode
from unified_planning.model.operators import OperatorKind
from typing import List, Tuple
from itertools import product


class Nnf:
    """Class used to transform a logic expression into the equivalent
    Negation Normal Form expression.

    This is done first by removing all the Implications and Equalities,
    then by pushing all the not to the leaves of the Tree representing the expression."""

    def __init__(self, env: "unified_planning.environment.Environment"):
        self.env = env
        self.manager = env.expression_manager

    def get_nnf_expression(self, expression: FNode) -> FNode:
        """Function used to transform a logic expression into the equivalent
        Negational Normal Form expression.

        This is done first by removing all the Implications and Equalities,
        and by pushing all the not to the leaves of the Tree representing the expression.

        For example, the form: !(a => (b && c)) becomes:
        a && (!b || !c), therefore it is a NNF."""
        stack: List[Tuple[bool, FNode, bool]] = []
        stack.append((True, expression, False))
        solved: List[FNode] = []
        while len(stack) > 0:
            p, e, status = stack.pop()
            if status:
                if e.is_and():
                    args = [solved.pop() for _ in range(len(e.args))]
                    if p:
                        new_e = self.manager.And(args)
                    else:
                        new_e = self.manager.Or(args)
                    solved.append(new_e)
                elif e.is_or():
                    args = [solved.pop() for _ in range(len(e.args))]
                    if p:
                        new_e = self.manager.Or(args)
                    else:
                        new_e = self.manager.And(args)
                    solved.append(new_e)
                else:
                    raise UPUnreachableCodeError(
                        "This code branch should never be reached!"
                    )
            else:
                if e.is_not():
                    stack.append((not p, e.arg(0), False))
                elif e.is_and() or e.is_or():
                    stack.append((p, e, True))
                    for arg in e.args:
                        stack.append((p, arg, False))
                elif e.is_implies():
                    na1 = self.manager.Not(e.arg(0))
                    new_e = self.manager.Or(na1, e.arg(1))
                    # stack.append((p, new_e, False)) would be enough.
                    # but this requires more iterations on the stack
                    # while the arguments can be expanded in this
                    # iteration.
                    stack.append((p, new_e, True))
                    stack.append((not p, e.arg(0), False))
                    stack.append((p, e.arg(1), False))
                elif e.is_iff():
                    na1 = self.manager.Not(e.arg(0))
                    na2 = self.manager.Not(e.arg(1))
                    e1 = self.manager.And(e.arg(0), e.arg(1))
                    e2 = self.manager.And(na1, na2)
                    new_e = self.manager.Or(e1, e2)
                    # stack.append((p, new_e, False)) would be enough.
                    # but this requires more iterations on the stack
                    # while the arguments can be expanded in this
                    # iteration.
                    stack.append((p, new_e, True))
                    stack.append((p, e1, True))
                    stack.append((p, e.arg(0), False))
                    stack.append((p, e.arg(1), False))
                    stack.append((p, e2, True))
                    stack.append((not p, e.arg(0), False))
                    stack.append((not p, e.arg(1), False))
                else:
                    if p:
                        solved.append(e)
                    else:
                        solved.append(self.manager.Not(e))
        assert len(solved) == 1  # sanity check
        return solved.pop()


class Dnf(walkers.dag.DagWalker):
    """Class used to transform a logic expression into the equivalent
    Disjunctive Normal Form expression.

    This is done first by transforming the expression into a NNF expression,
    and then every And and Or are propagated to be a unique equivalent Or of
    Ands or Atomic expressions, where 'atomic expressions' could also be a
    Not of an atomic expression.
    """

    def __init__(self, env: "unified_planning.environment.Environment"):
        walkers.dag.DagWalker.__init__(self, True)
        self.env = env
        self.manager = env.expression_manager
        self._nnf = Nnf(self.env)
        self._simplifier = walkers.simplifier.Simplifier(self.env)

    def get_dnf_expression(self, expression: FNode) -> FNode:
        """Function used to transform a logic expression into the equivalent
        Disjunctive Normal Form expression.

        This is done first by transforming the expression into a NNF expression,
        and then every And and Or are propagated to be a unique equivalent Or of
        Ands or Atomic expressions, where 'atomic expressions' could also be a
        Not of an atomic expression.

        For example, the form: !(a => (b && c)) becomes:
        a && (!b || !c), in NNF form, and then:
        (a && !b) || (a && !c), therefore a DNF expression."""
        nnf_exp = self._nnf.get_nnf_expression(expression)
        tuples = self.walk(nnf_exp)
        return self.manager.Or(self.manager.And(and_args) for and_args in tuples)

    def walk_and(
        self, expression: FNode, args: List[List[List[FNode]]], **kwargs
    ) -> List[List[FNode]]:
        res: List[List[FNode]] = []
        tuples = product(*args)
        # tuples is an iterable of tuples, where each tuple
        # represents one son of the resulting Or.
        # list will contain each son of the resulting And.
        # for example:
        #   tuples = ([a, b], [c]) ([d])
        # will result in
        #   Or(And(a, b, c), And(d))
        for conj_list in tuples:
            big_conjunction = [lit for conj in conj_list for lit in conj]
            simp = self._simplifier.simplify(self.manager.And(big_conjunction))
            if simp.is_true():
                return []
            elif simp.is_false():
                pass
            elif simp.is_and():
                res.append(simp.args)
            else:
                res.append([simp])
        return res

    def walk_or(
        self, expression: FNode, args: List[List[List[FNode]]], **kwargs
    ) -> List[List[FNode]]:
        return [conjunction for disjunction in args for conjunction in disjunction]

    @walkers.handles(set(OperatorKind) - set({OperatorKind.AND, OperatorKind.OR}))
    def walk_all(
        self, expression: FNode, args: List[List[List[FNode]]], **kwargs
    ) -> List[List[FNode]]:
        return [[expression]]



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""This module defines the negative preconditions remover class."""

import unified_planning as up
import unified_planning.engines as engines
from unified_planning.engines.mixins.compiler import CompilationKind, CompilerMixin
from unified_planning.engines.results import CompilerResult
from unified_planning.model import (
    Fluent,
    Problem,
    InstantaneousAction,
    DurativeAction,
    FNode,
    Action,
    Effect,
    Timing,
    ProblemKind,
)
from unified_planning.model.walkers.identitydag import IdentityDagWalker
from unified_planning.engines.compilers.utils import get_fresh_name, replace_action
from unified_planning.exceptions import (
    UPExpressionDefinitionError,
    UPProblemDefinitionError,
)
from typing import List, Dict, Union
from functools import partial


class NegativeFluentRemover(IdentityDagWalker):
    def __init__(self, problem, env):
        self._env = env
        IdentityDagWalker.__init__(self, self._env)
        self._fluent_mapping: Dict[Fluent, Fluent] = {}
        self._problem = problem

    def remove_negative_fluents(self, expression: FNode) -> FNode:
        return self.walk(expression)

    def walk_not(self, expression: FNode, args: List[FNode], **kwargs) -> FNode:
        assert len(args) == 1
        if not args[0].is_fluent_exp():
            raise UPExpressionDefinitionError(
                f"Expression: {expression} is not in NNF."
            )
        neg_fluent = self._fluent_mapping.get(args[0].fluent(), None)
        if neg_fluent is not None:
            return self._env.expression_manager.FluentExp(
                neg_fluent, tuple(args[0].args)
            )
        f = args[0].fluent()
        nf = Fluent(get_fresh_name(self._problem, f.name), f.type, f.signature, f._env)
        self._fluent_mapping[f] = nf
        return self._env.expression_manager.FluentExp(nf, tuple(args[0].args))

    @property
    def fluent_mapping(self) -> Dict[Fluent, Fluent]:
        return self._fluent_mapping


class NegativeConditionsRemover(engines.engine.Engine, CompilerMixin):
    """Negative conditions remover class: this class offers the capability
    to transform a problem with negative conditions into one
    without negative conditions.

    This is done by substituting every fluent that appears with a Not into the conditions
    with different fluent representing  his negation."""

    @property
    def name(self):
        return "ncrm"

    @staticmethod
    def supported_kind() -> ProblemKind:
        supported_kind = ProblemKind()
        supported_kind.set_problem_class("ACTION_BASED")
        supported_kind.set_typing("FLAT_TYPING")
        supported_kind.set_typing("HIERARCHICAL_TYPING")
        supported_kind.set_numbers("CONTINUOUS_NUMBERS")
        supported_kind.set_numbers("DISCRETE_NUMBERS")
        supported_kind.set_fluents_type("NUMERIC_FLUENTS")
        supported_kind.set_fluents_type("OBJECT_FLUENTS")
        supported_kind.set_conditions_kind("NEGATIVE_CONDITIONS")
        supported_kind.set_conditions_kind("DISJUNCTIVE_CONDITIONS")
        supported_kind.set_conditions_kind("EQUALITY")
        supported_kind.set_conditions_kind("EXISTENTIAL_CONDITIONS")
        supported_kind.set_conditions_kind("UNIVERSAL_CONDITIONS")
        supported_kind.set_effects_kind("CONDITIONAL_EFFECTS")
        supported_kind.set_effects_kind("INCREASE_EFFECTS")
        supported_kind.set_effects_kind("DECREASE_EFFECTS")
        supported_kind.set_time("CONTINUOUS_TIME")
        supported_kind.set_time("DISCRETE_TIME")
        supported_kind.set_time("INTERMEDIATE_CONDITIONS_AND_EFFECTS")
        supported_kind.set_time("TIMED_EFFECT")
        supported_kind.set_time("TIMED_GOALS")
        supported_kind.set_time("DURATION_INEQUALITIES")
        return supported_kind

    @staticmethod
    def supports(problem_kind):
        return problem_kind <= NegativeConditionsRemover.supported_kind()

    @staticmethod
    def supports_compilation(compilation_kind: CompilationKind) -> bool:
        return compilation_kind == CompilationKind.NEGATIVE_CONDITIONS_REMOVING

    def _compile(
        self,
        problem: "up.model.AbstractProblem",
        compilation_kind: "up.engines.CompilationKind",
    ) -> CompilerResult:
        assert isinstance(problem, Problem)

        env = problem.env
        simplifier = env.simplifier

        fluent_remover = NegativeFluentRemover(problem, env)

        new_to_old: Dict[Action, Action] = {}

        new_problem = Problem(f"{self.name}_{problem.name}", env)
        for o in problem.all_objects:
            new_problem.add_object(o)

        name_action_map: Dict[str, Union[InstantaneousAction, DurativeAction]] = {}
        for action in problem.actions:
            if isinstance(action, InstantaneousAction):
                new_action = action.clone()
                new_action.name = get_fresh_name(new_problem, action.name)
                new_action.clear_preconditions()
                for p in action.preconditions:
                    np = fluent_remover.remove_negative_fluents(p)
                    new_action.add_precondition(np)
                for ce in new_action.conditional_effects:
                    ce.set_condition(
                        fluent_remover.remove_negative_fluents(ce.condition)
                    )
                name_action_map[action.name] = new_action
            elif isinstance(action, DurativeAction):
                new_durative_action = action.clone()
                new_durative_action.name = get_fresh_name(new_problem, action.name)
                new_durative_action.clear_conditions()
                for i, cl in action.conditions.items():
                    for c in cl:
                        nc = fluent_remover.remove_negative_fluents(c)
                        new_durative_action.add_condition(i, nc)
                for t, cel in new_durative_action.conditional_effects.items():
                    for ce in cel:
                        ce.set_condition(
                            fluent_remover.remove_negative_fluents(ce.condition)
                        )
                name_action_map[action.name] = new_durative_action
            else:
                raise NotImplementedError

        for t, el in problem.timed_effects.items():
            for e in el:
                new_problem._add_effect_instance(t, e.clone())

        for t, el in new_problem.timed_effects.items():
            for e in el:
                if e.is_conditional():
                    e.set_condition(fluent_remover.remove_negative_fluents(e.condition))

        for i, gl in problem.timed_goals.items():
            for g in gl:
                ng = fluent_remover.remove_negative_fluents(g)
                new_problem.add_timed_goal(i, ng)

        for g in problem.goals:
            ng = fluent_remover.remove_negative_fluents(g)
            new_problem.add_goal(ng)

        # fluent_mapping is the map between a fluent and it's negation, when the
        # negation is None it means the fluent is never found in a negation into
        # every condititon analized before; therefore it does not need to exist.
        fluent_mapping = fluent_remover.fluent_mapping
        for f in problem.fluents:
            new_problem.add_fluent(f)
            fneg = fluent_mapping.get(f, None)
            if fneg is not None:
                new_problem.add_fluent(fneg)

        for fl, v in problem.initial_values.items():
            fneg = fluent_mapping.get(fl.fluent(), None)
            new_problem.set_initial_value(fl, v)
            if fneg is not None:
                if v.bool_constant_value():
                    new_problem.set_initial_value(
                        env.expression_manager.FluentExp(fneg, tuple(fl.args)),
                        env.expression_manager.FALSE(),
                    )
                else:
                    new_problem.set_initial_value(
                        env.expression_manager.FluentExp(fneg, tuple(fl.args)),
                        env.expression_manager.TRUE(),
                    )

        for action in problem.actions:
            if isinstance(action, InstantaneousAction):
                new_action = name_action_map[action.name]
                new_effects: List[Effect] = []
                for e in new_action.effects:
                    fl, v = e.fluent, e.value
                    fneg = fluent_mapping.get(fl.fluent(), None)
                    if fneg is not None:
                        simplified_not_v = simplifier.simplify(
                            env.expression_manager.Not(v)
                        )
                        new_effects.append(
                            Effect(
                                env.expression_manager.FluentExp(fneg, tuple(fl.args)),
                                simplified_not_v,
                                e.condition,
                                e.kind,
                            )
                        )
                for ne in new_effects:
                    new_action._add_effect_instance(ne)
                new_problem.add_action(new_action)
                new_to_old[new_action] = action
            elif isinstance(action, DurativeAction):
                new_durative_action = name_action_map[action.name]
                new_durative_action.set_duration_constraint(action.duration)

                for t, el in new_durative_action.effects.items():
                    for e in el:
                        fl, v = e.fluent, e.value
                        fneg = fluent_mapping.get(fl.fluent(), None)
                        if fneg is not None:
                            simplified_not_v = simplifier.simplify(
                                env.expression_manager.Not(v)
                            )
                            new_durative_action._add_effect_instance(
                                t,
                                Effect(
                                    env.expression_manager.FluentExp(
                                        fneg, tuple(fl.args)
                                    ),
                                    simplified_not_v,
                                    e.condition,
                                    e.kind,
                                ),
                            )
                new_problem.add_action(new_durative_action)
                new_to_old[new_durative_action] = action
            else:
                raise NotImplementedError

        for t, el in new_problem.timed_effects.items():
            for e in el:
                fl, v = e.fluent, e.value
                fneg = fluent_mapping.get(fl.fluent(), None)
                if fneg is not None:
                    simplified_not_v = simplifier.simplify(
                        env.expression_manager.Not(v)
                    )
                    new_problem._add_effect_instance(
                        t,
                        Effect(
                            env.expression_manager.FluentExp(fneg, tuple(fl.args)),
                            simplified_not_v,
                            e.condition,
                            e.kind,
                        ),
                    )

        return CompilerResult(
            new_problem, partial(replace_action, map=new_to_old), self.name
        )



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import unified_planning
from unified_planning.environment import get_env
from unified_planning.shortcuts import *
from unified_planning.model.problem_kind import (
    classical_kind,
    full_classical_kind,
    basic_temporal_kind,
    full_numeric_kind,
    hierarchical_kind,
)
from unified_planning.test import (
    TestCase,
    skipIfNoPlanValidatorForProblemKind,
    skipIfNoOneshotPlannerForProblemKind,
)
from unified_planning.test.examples import get_example_problems
from unified_planning.engines import CompilationKind
from unified_planning.engines.compilers import QuantifiersRemover


class TestQuantifiersRemover(TestCase):
    def setUp(self):
        TestCase.setUp(self)
        self.problems = get_example_problems()

    @skipIfNoOneshotPlannerForProblemKind(classical_kind)
    @skipIfNoPlanValidatorForProblemKind(full_classical_kind)
    def test_basic_exists(self):
        problem = self.problems["basic_exists"].problem
        with Compiler(
            problem_kind=problem.kind,
            compilation_kind=CompilationKind.QUANTIFIERS_REMOVING,
        ) as qr:
            res = qr.compile(problem, CompilationKind.QUANTIFIERS_REMOVING)
            res_2 = qr.compile(problem, CompilationKind.QUANTIFIERS_REMOVING)
        uq_problem = res.problem
        uq_problem_2 = res_2.problem
        self.assertEqual(uq_problem, uq_problem_2)
        self.assertTrue(problem.kind.has_existential_conditions())
        self.assertFalse(uq_problem.kind.has_existential_conditions())
        self.assertEqual(len(problem.actions), len(uq_problem.actions))

        with OneshotPlanner(problem_kind=uq_problem.kind) as planner:
            self.assertNotEqual(planner, None)
            uq_plan = planner.solve(uq_problem).plan
            new_plan = uq_plan.replace_action_instances(res.map_back_action_instance)
            with PlanValidator(
                problem_kind=problem.kind, plan_kind=new_plan.kind
            ) as pv:
                self.assertTrue(pv.validate(problem, new_plan))

    @skipIfNoOneshotPlannerForProblemKind(classical_kind)
    @skipIfNoPlanValidatorForProblemKind(full_classical_kind)
    def test_basic_forall(self):
        problem = self.problems["basic_forall"].problem
        qr = QuantifiersRemover()
        res = qr.compile(problem, CompilationKind.QUANTIFIERS_REMOVING)
        uq_problem = res.problem
        self.assertTrue(problem.kind.has_universal_conditions())
        self.assertFalse(uq_problem.kind.has_universal_conditions())
        self.assertEqual(len(problem.actions), len(uq_problem.actions))

        with OneshotPlanner(problem_kind=uq_problem.kind) as planner:
            self.assertNotEqual(planner, None)
            uq_plan = planner.solve(uq_problem).plan
            new_plan = uq_plan.replace_action_instances(res.map_back_action_instance)
            with PlanValidator(
                problem_kind=problem.kind, plan_kind=new_plan.kind
            ) as pv:
                self.assertTrue(pv.validate(problem, new_plan))

    @skipIfNoOneshotPlannerForProblemKind(classical_kind.union(full_numeric_kind))
    @skipIfNoPlanValidatorForProblemKind(full_classical_kind.union(full_numeric_kind))
    def test_robot_locations_connected(self):
        problem = self.problems["robot_locations_connected"].problem
        qr = QuantifiersRemover()
        res = qr.compile(problem, CompilationKind.QUANTIFIERS_REMOVING)
        uq_problem = res.problem
        self.assertTrue(problem.kind.has_existential_conditions())
        self.assertFalse(uq_problem.kind.has_existential_conditions())
        self.assertEqual(len(problem.actions), len(uq_problem.actions))

        with OneshotPlanner(problem_kind=uq_problem.kind) as planner:
            self.assertNotEqual(planner, None)
            uq_plan = planner.solve(uq_problem).plan
            new_plan = uq_plan.replace_action_instances(res.map_back_action_instance)
            with PlanValidator(
                problem_kind=problem.kind, plan_kind=new_plan.kind
            ) as pv:
                self.assertTrue(pv.validate(problem, new_plan))

    @skipIfNoOneshotPlannerForProblemKind(classical_kind.union(full_numeric_kind))
    @skipIfNoPlanValidatorForProblemKind(full_classical_kind.union(full_numeric_kind))
    def test_robot_locations_visited(self):
        problem = self.problems["robot_locations_visited"].problem
        qr = QuantifiersRemover()
        res = qr.compile(problem, CompilationKind.QUANTIFIERS_REMOVING)
        uq_problem = res.problem
        self.assertTrue(problem.kind.has_existential_conditions())
        self.assertFalse(uq_problem.kind.has_existential_conditions())
        self.assertTrue(problem.kind.has_universal_conditions())
        self.assertFalse(uq_problem.kind.has_universal_conditions())
        self.assertEqual(len(problem.actions), len(uq_problem.actions))

        with OneshotPlanner(problem_kind=uq_problem.kind) as planner:
            self.assertNotEqual(planner, None)
            uq_plan = planner.solve(uq_problem).plan
            new_plan = uq_plan.replace_action_instances(res.map_back_action_instance)
            with PlanValidator(
                problem_kind=problem.kind, plan_kind=new_plan.kind
            ) as pv:
                self.assertTrue(pv.validate(problem, new_plan))

    def test_hierarchical_blocks_world_exists(self):
        problem = self.problems["hierarchical_blocks_world_exists"].problem
        qr = QuantifiersRemover()
        res = qr.compile(problem, CompilationKind.QUANTIFIERS_REMOVING)
        uq_problem = res.problem
        self.assertTrue(problem.kind.has_existential_conditions())
        self.assertFalse(uq_problem.kind.has_existential_conditions())
        self.assertTrue(uq_problem.kind.has_disjunctive_conditions())
        self.assertFalse(problem.kind.has_disjunctive_conditions())
        self.assertIn(
            "(on(block_1, block_1) or on(block_2, block_1) or on(block_3, block_1))",
            str(uq_problem.goals),
        )

    @skipIfNoOneshotPlannerForProblemKind(classical_kind.union(basic_temporal_kind))
    def test_timed_connected_locations(self):
        problem = self.problems["timed_connected_locations"].problem
        qr = QuantifiersRemover()
        res = qr.compile(problem, CompilationKind.QUANTIFIERS_REMOVING)
        uq_problem = res.problem
        self.assertTrue(problem.kind.has_existential_conditions())
        self.assertFalse(uq_problem.kind.has_existential_conditions())
        self.assertEqual(len(problem.actions), len(uq_problem.actions))

        with OneshotPlanner(problem_kind=uq_problem.kind) as planner:
            self.assertNotEqual(planner, None)
            uq_plan = planner.solve(uq_problem).plan
            new_plan = uq_plan.replace_action_instances(res.map_back_action_instance)
            for (s, a, d), (s_1, a_1, d_1) in zip(
                new_plan.timed_actions, uq_plan.timed_actions
            ):
                self.assertEqual(s, s_1)
                self.assertEqual(d, d_1)
                self.assertIn(a.action, problem.actions)

    def test_ad_hoc_1(self):
        Obj = UserType("Obj")
        x = Fluent("x")
        y = Fluent("y", BoolType(), obj=Obj)
        o = Variable("o", Obj)
        a = InstantaneousAction("a")
        o1 = Object("o1", Obj)
        o2 = Object("o2", Obj)
        o3 = Object("o3", Obj)
        a.add_effect(x, True, Exists(FluentExp(y, [o]), o))
        da = DurativeAction("da")
        da.add_effect(StartTiming(), x, True, Forall(FluentExp(y, [o]), o))
        problem = Problem("ad_hoc")
        problem.add_fluent(x)
        problem.add_fluent(y)
        problem.add_action(a)
        problem.add_action(da)
        problem.add_object(o1)
        problem.add_object(o2)
        problem.add_object(o3)
        problem.add_timed_effect(
            GlobalStartTiming(4),
            x,
            Forall(FluentExp(y, [o]), o),
            Exists(FluentExp(y, [o]), o),
        )
        problem.add_timed_goal(GlobalStartTiming(6), x)
        problem.add_timed_goal(
            OpenTimeInterval(GlobalStartTiming(8), GlobalStartTiming(10)), x
        )
        problem.set_initial_value(x, False)
        problem.set_initial_value(y(o1), True)
        problem.set_initial_value(y(o2), False)
        problem.set_initial_value(y(o3), True)
        problem.add_goal(x)
        qr = QuantifiersRemover()
        res = qr.compile(problem, CompilationKind.QUANTIFIERS_REMOVING)
        unq_problem = res.problem
        self.assertTrue(problem.kind.has_existential_conditions())
        self.assertTrue(problem.kind.has_universal_conditions())
        self.assertFalse(unq_problem.kind.has_existential_conditions())
        self.assertFalse(unq_problem.kind.has_universal_conditions())



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""This module defines the quantifiers remover class."""


import unified_planning as up
import unified_planning.engines as engines
from unified_planning.engines.mixins.compiler import CompilationKind, CompilerMixin
from unified_planning.engines.results import CompilerResult
from unified_planning.model import (
    Problem,
    InstantaneousAction,
    DurativeAction,
    Action,
    ProblemKind,
)
from unified_planning.model.walkers import ExpressionQuantifiersRemover
from unified_planning.engines.compilers.utils import get_fresh_name, replace_action
from typing import Dict
from functools import partial


class QuantifiersRemover(engines.engine.Engine, CompilerMixin):
    """Quantifiers remover class: this class offers the capability
    to transform a problem with quantifiers into a problem without.
    Every quantifier is compiled away by it's respective logic formula
    instantiated on object.
    For example the formula:
        Forall (s-sempahore) is_green(s)
    in a problem with 3 objects of type semaphores {s1, s2, s3} is compiled in:
        And(is_green(s1), is_green(s2), is_green(s3)).
    While the respective formula on the same problem:
        Exists (s-semaphore) is_green(s)
    becomes:
        Or(is_green(s1), is_green(s2), is_green(s3))."""

    @property
    def name(self):
        return "qurm"

    @staticmethod
    def supported_kind() -> ProblemKind:
        supported_kind = ProblemKind()
        supported_kind.set_problem_class("ACTION_BASED")
        supported_kind.set_typing("FLAT_TYPING")
        supported_kind.set_typing("HIERARCHICAL_TYPING")
        supported_kind.set_numbers("CONTINUOUS_NUMBERS")
        supported_kind.set_numbers("DISCRETE_NUMBERS")
        supported_kind.set_fluents_type("NUMERIC_FLUENTS")
        supported_kind.set_fluents_type("OBJECT_FLUENTS")
        supported_kind.set_conditions_kind("NEGATIVE_CONDITIONS")
        supported_kind.set_conditions_kind("DISJUNCTIVE_CONDITIONS")
        supported_kind.set_conditions_kind("EQUALITY")
        supported_kind.set_conditions_kind("EXISTENTIAL_CONDITIONS")
        supported_kind.set_conditions_kind("UNIVERSAL_CONDITIONS")
        supported_kind.set_effects_kind("CONDITIONAL_EFFECTS")
        supported_kind.set_effects_kind("INCREASE_EFFECTS")
        supported_kind.set_effects_kind("DECREASE_EFFECTS")
        supported_kind.set_time("CONTINUOUS_TIME")
        supported_kind.set_time("DISCRETE_TIME")
        supported_kind.set_time("INTERMEDIATE_CONDITIONS_AND_EFFECTS")
        supported_kind.set_time("TIMED_EFFECT")
        supported_kind.set_time("TIMED_GOALS")
        supported_kind.set_time("DURATION_INEQUALITIES")
        supported_kind.set_simulated_entities("SIMULATED_EFFECTS")
        return supported_kind

    @staticmethod
    def supports(problem_kind):
        return problem_kind <= QuantifiersRemover.supported_kind()

    @staticmethod
    def supports_compilation(compilation_kind: CompilationKind) -> bool:
        return compilation_kind == CompilationKind.QUANTIFIERS_REMOVING

    def _compile(
        self,
        problem: "up.model.AbstractProblem",
        compilation_kind: "up.engines.CompilationKind",
    ) -> CompilerResult:
        assert isinstance(problem, Problem)

        env = problem.env
        expression_quantifier_remover = ExpressionQuantifiersRemover(env)

        new_to_old: Dict[Action, Action] = {}

        new_problem = problem.clone()
        new_problem.name = f"{self.name}_{problem.name}"
        new_problem.clear_timed_goals()
        new_problem.clear_goals()
        for action in new_problem.actions:
            if isinstance(action, InstantaneousAction):
                original_action = problem.action(action.name)
                assert isinstance(original_action, InstantaneousAction)
                action.name = get_fresh_name(new_problem, action.name)
                action.clear_preconditions()
                for p in original_action.preconditions:
                    action.add_precondition(
                        expression_quantifier_remover.remove_quantifiers(p, problem)
                    )
                for e in action.effects:
                    if e.is_conditional():
                        e.set_condition(
                            expression_quantifier_remover.remove_quantifiers(
                                e.condition, problem
                            )
                        )
                    e.set_value(
                        expression_quantifier_remover.remove_quantifiers(
                            e.value, problem
                        )
                    )
                new_to_old[action] = original_action
            elif isinstance(action, DurativeAction):
                original_action = problem.action(action.name)
                assert isinstance(original_action, DurativeAction)
                action.name = get_fresh_name(new_problem, action.name)
                action.clear_conditions()
                for i, cl in original_action.conditions.items():
                    for c in cl:
                        action.add_condition(
                            i,
                            expression_quantifier_remover.remove_quantifiers(
                                c, problem
                            ),
                        )
                for t, el in action.effects.items():
                    for e in el:
                        if e.is_conditional():
                            e.set_condition(
                                expression_quantifier_remover.remove_quantifiers(
                                    e.condition, problem
                                )
                            )
                        e.set_value(
                            expression_quantifier_remover.remove_quantifiers(
                                e.value, problem
                            )
                        )
                new_to_old[action] = original_action
            else:
                raise NotImplementedError
        for t, el in new_problem.timed_effects.items():
            for e in el:
                if e.is_conditional():
                    e.set_condition(
                        expression_quantifier_remover.remove_quantifiers(
                            e.condition, problem
                        )
                    )
                e.set_value(
                    expression_quantifier_remover.remove_quantifiers(e.value, problem)
                )
        for i, gl in problem.timed_goals.items():
            for g in gl:
                ng = expression_quantifier_remover.remove_quantifiers(g, problem)
                new_problem.add_timed_goal(i, ng)
        for g in problem.goals:
            ng = expression_quantifier_remover.remove_quantifiers(g, problem)
            new_problem.add_goal(ng)

        return CompilerResult(
            new_problem, partial(replace_action, map=new_to_old), self.name
        )



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from fractions import Fraction
import os
from unified_planning.plans import ActionInstance
import unified_planning
from unified_planning.environment import get_env
from unified_planning.shortcuts import *
from unified_planning.test import (
    TestCase,
    main,
    skipIfNoPlanValidatorForProblemKind,
    skipIfNoOneshotPlannerForProblemKind,
)
from unified_planning.test.examples import get_example_problems
from unified_planning.model.problem_kind import (
    basic_classical_kind,
    classical_kind,
    basic_temporal_kind,
    full_classical_kind,
)
from unified_planning.engines.compilers import NegativeConditionsRemover
from unified_planning.engines import SequentialPlanValidator as PV, CompilationKind
from unified_planning.exceptions import (
    UPExpressionDefinitionError,
    UPProblemDefinitionError,
)


class TestNegativeConditionsRemover(TestCase):
    def setUp(self):
        TestCase.setUp(self)
        self.env = get_env()
        self.problems = get_example_problems()

    @skipIfNoOneshotPlannerForProblemKind(basic_classical_kind)
    @skipIfNoPlanValidatorForProblemKind(classical_kind)
    def test_basic(self):
        problem = self.problems["basic"].problem
        with Compiler(
            problem_kind=problem.kind,
            compilation_kind=CompilationKind.NEGATIVE_CONDITIONS_REMOVING,
        ) as npr:
            res = npr.compile(problem, CompilationKind.NEGATIVE_CONDITIONS_REMOVING)
        positive_problem = res.problem
        self.assertEqual(len(problem.fluents) + 1, len(positive_problem.fluents))
        self.assertTrue(problem.kind.has_negative_conditions())
        self.assertFalse(positive_problem.kind.has_negative_conditions())
        with OneshotPlanner(problem_kind=positive_problem.kind) as planner:
            self.assertNotEqual(planner, None)
            positive_plan = planner.solve(positive_problem).plan
            new_plan = positive_plan.replace_action_instances(
                res.map_back_action_instance
            )
            with PlanValidator(
                problem_kind=problem.kind, plan_kind=new_plan.kind
            ) as PV:
                self.assertTrue(PV.validate(problem, new_plan))

    @skipIfNoOneshotPlannerForProblemKind(basic_classical_kind)
    @skipIfNoPlanValidatorForProblemKind(classical_kind)
    def test_robot_loader_mod(self):
        problem = self.problems["robot_loader_mod"].problem
        npr = NegativeConditionsRemover()
        res = npr.compile(problem, CompilationKind.NEGATIVE_CONDITIONS_REMOVING)
        positive_problem = res.problem
        res_2 = npr.compile(problem, CompilationKind.NEGATIVE_CONDITIONS_REMOVING)
        positive_problem_2 = res_2.problem
        self.assertEqual(positive_problem, positive_problem_2)
        self.assertEqual(len(problem.fluents) + 4, len(positive_problem.fluents))
        self.assertTrue(problem.kind.has_negative_conditions())
        self.assertFalse(positive_problem.kind.has_negative_conditions())
        with OneshotPlanner(problem_kind=positive_problem.kind) as planner:
            self.assertNotEqual(planner, None)
            positive_plan = planner.solve(positive_problem).plan
            new_plan = positive_plan.replace_action_instances(
                res.map_back_action_instance
            )
            with PlanValidator(
                problem_kind=problem.kind, plan_kind=new_plan.kind
            ) as PV:
                self.assertTrue(PV.validate(problem, new_plan))

    @skipIfNoOneshotPlannerForProblemKind(
        basic_classical_kind.union(basic_temporal_kind)
    )
    @skipIfNoPlanValidatorForProblemKind(classical_kind.union(basic_temporal_kind))
    def test_matchcellar(self):
        problem = self.problems["matchcellar"].problem
        npr = NegativeConditionsRemover()
        res = npr.compile(problem, CompilationKind.NEGATIVE_CONDITIONS_REMOVING)
        positive_problem = res.problem
        self.assertTrue(problem.kind.has_negative_conditions())
        self.assertFalse(positive_problem.kind.has_negative_conditions())
        self.assertTrue(problem.kind.has_negative_conditions())
        self.assertFalse(positive_problem.kind.has_negative_conditions())
        with OneshotPlanner(problem_kind=positive_problem.kind) as planner:
            self.assertNotEqual(planner, None)
            positive_plan = planner.solve(positive_problem).plan
            new_plan = positive_plan.replace_action_instances(
                res.map_back_action_instance
            )
            with PlanValidator(
                problem_kind=problem.kind, plan_kind=new_plan.kind
            ) as PV:
                self.assertTrue(PV.validate(problem, new_plan))
        self.assertEqual(len(problem.fluents) + 1, len(positive_problem.fluents))
        light_match = problem.action("light_match")
        mend_fuse = problem.action("mend_fuse")
        m1 = problem.object("m1")
        m2 = problem.object("m2")
        m3 = problem.object("m3")
        f1 = problem.object("f1")
        f2 = problem.object("f2")
        f3 = problem.object("f3")
        light_m1 = ActionInstance(light_match, (ObjectExp(m1),))
        light_m2 = ActionInstance(light_match, (ObjectExp(m2),))
        light_m3 = ActionInstance(light_match, (ObjectExp(m3),))
        mend_f1 = ActionInstance(mend_fuse, (ObjectExp(f1),))
        mend_f2 = ActionInstance(mend_fuse, (ObjectExp(f2),))
        mend_f3 = ActionInstance(mend_fuse, (ObjectExp(f3),))
        expected_ai = [light_m1, light_m2, light_m3, mend_f1, mend_f2, mend_f3]
        self.assertEqual(len(new_plan.timed_actions), len(expected_ai))
        for ai in expected_ai:
            self.assertIn(ai, new_plan)

    @skipIfNoOneshotPlannerForProblemKind(full_classical_kind)
    @skipIfNoPlanValidatorForProblemKind(full_classical_kind)
    def test_basic_conditional(self):
        problem = self.problems["basic_conditional"].problem
        npr = NegativeConditionsRemover()
        res = npr.compile(problem, CompilationKind.NEGATIVE_CONDITIONS_REMOVING)
        positive_problem = res.problem
        self.assertEqual(len(problem.fluents) + 2, len(positive_problem.fluents))
        self.assertTrue(problem.kind.has_negative_conditions())
        self.assertFalse(positive_problem.kind.has_negative_conditions())
        with OneshotPlanner(problem_kind=positive_problem.kind) as planner:
            self.assertNotEqual(planner, None)
            positive_plan = planner.solve(positive_problem).plan
            new_plan = positive_plan.replace_action_instances(
                res.map_back_action_instance
            )
            with PlanValidator(
                problem_kind=problem.kind, plan_kind=new_plan.kind
            ) as PV:
                self.assertTrue(PV.validate(problem, new_plan))

    def test_temporal_conditional(self):
        problem = self.problems["temporal_conditional"].problem
        npr = NegativeConditionsRemover()
        res = npr.compile(problem, CompilationKind.NEGATIVE_CONDITIONS_REMOVING)
        positive_problem = res.problem
        self.assertEqual(len(problem.fluents) + 3, len(positive_problem.fluents))
        self.assertTrue(problem.kind.has_negative_conditions())
        self.assertFalse(positive_problem.kind.has_negative_conditions())

    def test_ad_hoc_1(self):
        x = Fluent("x")
        y = Fluent("y")
        a = InstantaneousAction("a")
        a.add_precondition(And(Not(x), Not(y)))
        a.add_effect(x, True)
        problem = Problem("ad_hoc")
        problem.add_fluent(x)
        problem.add_fluent(y)
        problem.add_action(a)
        problem.set_initial_value(x, False)
        problem.set_initial_value(y, False)
        problem.add_goal(x)
        problem.add_goal(Not(y))
        problem.add_goal(Not(Iff(x, y)))
        problem.add_timed_goal(GlobalStartTiming(5), x)
        problem.add_timed_goal(
            ClosedTimeInterval(GlobalStartTiming(3), GlobalStartTiming(4)), x
        )
        npr = NegativeConditionsRemover()
        with self.assertRaises(UPExpressionDefinitionError) as e:
            res = npr.compile(problem, CompilationKind.NEGATIVE_CONDITIONS_REMOVING)
        self.assertIn(f"Expression: {Not(Iff(x, y))} is not in NNF.", str(e.exception))

    def test_ad_hoc_2(self):
        x = Fluent("x")
        y = Fluent("y")
        t = GlobalStartTiming(5)
        problem = Problem("ad_hoc")
        problem.add_fluent(x)
        problem.add_fluent(y)
        problem.add_timed_effect(t, y, x, Not(y))
        problem.set_initial_value(x, True)
        problem.set_initial_value(y, False)
        problem.add_goal(x)
        npr = NegativeConditionsRemover()
        res = npr.compile(problem, CompilationKind.NEGATIVE_CONDITIONS_REMOVING)
        positive_problem = res.problem
        self.assertEqual(len(problem.fluents) + 1, len(positive_problem.fluents))
        y__negated__ = Fluent("y_0")
        test_problem = Problem(positive_problem.name)
        test_problem.add_fluent(x)
        test_problem.add_fluent(y)
        test_problem.add_fluent(y__negated__)
        test_problem.add_timed_effect(t, y, x, y__negated__)
        test_problem.add_timed_effect(t, y__negated__, Not(x), y__negated__)
        test_problem.set_initial_value(x, True)
        test_problem.set_initial_value(y, False)
        test_problem.set_initial_value(y__negated__, True)
        test_problem.add_goal(x)
        self.assertEqual(positive_problem, test_problem)



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import subprocess
import re
from typing import Any, Tuple, Union, TYPE_CHECKING

from unified_planning.environment import Environment


if TYPE_CHECKING:
    # can't actually subclass Any at runtime
    AnyBaseClass = Any
else:
    AnyBaseClass = object


VERSION: Tuple[Union[int, str], ...] = (0, 4, 2)
__version__ = ".".join(str(x) for x in VERSION)

# Try to provide human-readable version of latest commit for dev versions
# E.g. v0.5.1-4-g49a49f2-wip
#      * 4 commits after tag v0.5.1
#      * Latest commit "49a49f2"
#      * -wip: Working tree is dirty (non committed stuff)
# See: https://git-scm.com/docs/git-describe
try:
    git_version = subprocess.check_output(
        ["git", "describe", "--tags", "--dirty=-wip"], stderr=subprocess.STDOUT
    )
    output = git_version.strip().decode("ascii")
    data = output.split("-")
    tag = data[0]
    match = re.match(r"^v(\d+)\.(\d)+\.(\d)$", tag)
    if match is not None:
        MAJOR, MINOR, REL = tuple(int(x) for x in match.groups())

    try:
        COMMITS = int(data[1])
    except ValueError:
        COMMITS = 0

    if data[-1] == "wip":
        if COMMITS == 0:
            VERSION = (MAJOR, MINOR, REL, "post", 1)
            __version__ = f"{MAJOR}.{MINOR}.{REL}.post1"
        else:
            VERSION = (MAJOR, MINOR, REL, COMMITS, "post", 1)
            __version__ = f"{MAJOR}.{MINOR}.{REL}.{COMMITS}.post1"
    else:
        VERSION = (MAJOR, MINOR, REL, COMMITS, "dev", 1)
        __version__ = f"{MAJOR}.{MINOR}.{REL}.{COMMITS}.dev1"
except Exception as ex:
    pass



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from unified_planning.shortcuts import *
from unified_planning.test import TestCase, main, skipIfEngineNotAvailable
from unified_planning.test.examples import get_example_problems
from unified_planning.io import ANMLWriter
import tempfile
import os


class TestPythonWriter(TestCase):
    def setUp(self):
        TestCase.setUp(self)
        self.problems = get_example_problems()

    def test_basic(self):
        problem = self.problems["basic"].problem
        aw = ANMLWriter(problem)
        anml_problem = aw.get_problem()
        self.assertIn("fluent boolean x;\n", anml_problem)
        self.assertIn("action a() {\n", anml_problem)
        self.assertIn("   [ start ] (not x);\n", anml_problem)
        self.assertIn("   [ start ] x := true;\n", anml_problem)
        self.assertIn("};\n", anml_problem)
        self.assertIn("[ start ] x := false;\n", anml_problem)
        self.assertIn("[ end ] x;\n", anml_problem)
        self.assertEqual(anml_problem.count("\n"), 7)

    def test_hierarchical_blocks_world(self):
        problem = self.problems["hierarchical_blocks_world"].problem
        aw = ANMLWriter(problem)
        anml_problem = aw.get_problem()
        self.assertIn("type Entity;\n", anml_problem)
        self.assertIn("type Location < Entity;\n", anml_problem)
        self.assertIn("type Movable < Location;\n", anml_problem)
        self.assertIn("type Unmovable < Location;\n", anml_problem)
        self.assertIn("type TableSpace < Unmovable;\n", anml_problem)
        self.assertIn("type Block < Movable;\n", anml_problem)
        self.assertIn("fluent boolean clear(Location space);\n", anml_problem)
        self.assertIn(
            "fluent boolean on(Movable object_, Location space);\n", anml_problem
        )
        self.assertIn(
            "action move(Movable item, Location l_from, Location l_to) {\n",
            anml_problem,
        )
        self.assertIn("   [ start ] clear(item);\n", anml_problem)
        self.assertIn("   [ start ] clear(l_to);\n", anml_problem)
        self.assertIn("   [ start ] on(item, l_from);\n", anml_problem)
        self.assertIn("   [ start ] clear(l_from) := true;\n", anml_problem)
        self.assertIn("   [ start ] on(item, l_from) := false;\n", anml_problem)
        self.assertIn("   [ start ] clear(l_to) := false;\n", anml_problem)
        self.assertIn("[ start ] on(item, l_to) := true;\n", anml_problem)
        self.assertIn("};\n", anml_problem)
        self.assertIn("instance TableSpace ts_1, ts_2, ts_3;\n", anml_problem)
        self.assertIn("instance Block block_1, block_2, block_3;\n", anml_problem)
        self.assertIn("[ start ] clear(ts_2) := true;\n", anml_problem)
        self.assertIn("[ start ] clear(ts_3) := true;\n", anml_problem)
        self.assertIn("[ start ] clear(block_2) := true;\n", anml_problem)
        self.assertIn("[ start ] on(block_3, ts_1) := true;\n", anml_problem)
        self.assertIn("[ start ] on(block_1, block_3) := true;\n", anml_problem)
        self.assertIn("[ start ] on(block_2, block_1) := true;\n", anml_problem)
        self.assertIn("[ start ] clear(ts_1) := false;\n", anml_problem)
        self.assertIn("[ start ] clear(block_1) := false;\n", anml_problem)
        self.assertIn("[ start ] clear(block_3) := false;\n", anml_problem)
        self.assertIn("[ start ] on(block_1, ts_1) := false;\n", anml_problem)
        self.assertIn("[ start ] on(block_2, ts_1) := false;\n", anml_problem)
        self.assertIn("[ start ] on(block_1, ts_2) := false;\n", anml_problem)
        self.assertIn("[ start ] on(block_2, ts_2) := false;\n", anml_problem)
        self.assertIn("[ start ] on(block_3, ts_2) := false;\n", anml_problem)
        self.assertIn("[ start ] on(block_1, ts_3) := false;\n", anml_problem)
        self.assertIn("[ start ] on(block_2, ts_3) := false;\n", anml_problem)
        self.assertIn("[ start ] on(block_3, ts_3) := false;\n", anml_problem)
        self.assertIn("[ start ] on(block_1, block_1) := false;\n", anml_problem)
        self.assertIn("[ start ] on(block_3, block_1) := false;\n", anml_problem)
        self.assertIn("[ start ] on(block_1, block_2) := false;\n", anml_problem)
        self.assertIn("[ start ] on(block_2, block_2) := false;\n", anml_problem)
        self.assertIn("[ start ] on(block_3, block_2) := false;\n", anml_problem)
        self.assertIn("[ start ] on(block_2, block_3) := false;\n", anml_problem)
        self.assertIn("[ start ] on(block_3, block_3) := false;\n", anml_problem)
        self.assertIn("[ end ] on(block_1, ts_3);\n", anml_problem)
        self.assertIn("[ end ] on(block_2, block_1);\n", anml_problem)
        self.assertIn("[ end ] on(block_3, block_2);\n", anml_problem)
        self.assertEqual(anml_problem.count("\n"), 46)

    def test_timed_connected_locations(self):
        problem = self.problems["timed_connected_locations"].problem
        aw = ANMLWriter(problem)
        anml_problem = aw.get_problem()
        expected_result = """type Location;
fluent boolean is_at(Location position);
constant boolean is_connected(Location location_1, Location location_2);
action move(Location l_from, Location l_to) {
   duration >= 6 and duration <= 6;
   [ start, start ] is_at(l_from);
   [ start, start ] (not is_at(l_to));
   [ start, start ] (exists(Location mid_loc) { ((not ((mid_loc == l_from) or (mid_loc == l_to))) and (is_connected(l_from, mid_loc) or is_connected(mid_loc, l_from)) and (is_connected(l_to, mid_loc) or is_connected(mid_loc, l_to))) });
   [ start, end ] (exists(Location mid_loc) { ((not ((mid_loc == l_from) or (mid_loc == l_to))) and (is_connected(l_from, mid_loc) or is_connected(mid_loc, l_from)) and (is_connected(l_to, mid_loc) or is_connected(mid_loc, l_to))) });
   [ start + 1 ] is_at(l_from) := false;
   [ end + 5 ] is_at(l_to) := true;
};
instance Location l1, l2, l3, l4, l5;
[ start ] is_at(l1) := true;
is_connected(l1, l2) := true;
is_connected(l2, l3) := true;
is_connected(l3, l4) := true;
is_connected(l4, l5) := true;
[ start ] is_at(l2) := false;
[ start ] is_at(l3) := false;
[ start ] is_at(l4) := false;
[ start ] is_at(l5) := false;
is_connected(l1, l1) := false;
is_connected(l2, l1) := false;
is_connected(l3, l1) := false;
is_connected(l4, l1) := false;
is_connected(l5, l1) := false;
is_connected(l2, l2) := false;
is_connected(l3, l2) := false;
is_connected(l4, l2) := false;
is_connected(l5, l2) := false;
is_connected(l1, l3) := false;
is_connected(l3, l3) := false;
is_connected(l4, l3) := false;
is_connected(l5, l3) := false;
is_connected(l1, l4) := false;
is_connected(l2, l4) := false;
is_connected(l4, l4) := false;
is_connected(l5, l4) := false;
is_connected(l1, l5) := false;
is_connected(l2, l5) := false;
is_connected(l3, l5) := false;
is_connected(l5, l5) := false;
[ end ] is_at(l5);
"""
        self.assertEqual(anml_problem, expected_result)

    def test_matchcellar(self):
        problem = self.problems["matchcellar"].problem
        aw = ANMLWriter(problem)
        anml_problem = aw.get_problem()
        expected_result = """type Match;
type Fuse;
fluent boolean handfree;
fluent boolean light;
fluent boolean match_used(Match match);
fluent boolean fuse_mended(Fuse fuse);
action light_match(Match m) {
   duration >= 6 and duration <= 6;
   [ start, start ] (not match_used(m));
   [ start ] match_used(m) := true;
   [ start ] light := true;
   [ end ] light := false;
};
action mend_fuse(Fuse f) {
   duration >= 5 and duration <= 5;
   [ start, start ] handfree;
   [ start, end ] light;
   [ start ] handfree := false;
   [ end ] fuse_mended(f) := true;
   [ end ] handfree := true;
};
instance Match m1, m2, m3;
instance Fuse f1, f2, f3;
[ start ] light := false;
[ start ] handfree := true;
[ start ] match_used(m1) := false;
[ start ] match_used(m2) := false;
[ start ] match_used(m3) := false;
[ start ] fuse_mended(f1) := false;
[ start ] fuse_mended(f2) := false;
[ start ] fuse_mended(f3) := false;
[ end ] fuse_mended(f1);
[ end ] fuse_mended(f2);
[ end ] fuse_mended(f3);
"""
        self.assertEqual(anml_problem, expected_result)

    def test_robot_fluent_of_user_type(self):
        problem = self.problems["robot_fluent_of_user_type"].problem
        aw = ANMLWriter(problem)
        anml_problem = aw.get_problem()
        expected_result = """type Location;
type Robot;
fluent Location is_at(Robot robot);
action move(Robot robot, Location l_from, Location l_to) {
   [ start ] (is_at(robot) == l_from);
   [ start ] (not (is_at(robot) == l_to));
   [ start ] is_at(robot) := l_to;
};
instance Location l1, l2;
instance Robot r1, r2;
[ start ] is_at(r1) := l2;
[ start ] is_at(r2) := l1;
[ end ] (is_at(r1) == l1);
[ end ] (is_at(r2) == l2);
"""
        self.assertEqual(anml_problem, expected_result)

    def test_robot(self):
        problem = self.problems["robot"].problem
        aw = ANMLWriter(problem)
        anml_problem = aw.get_problem()
        expected_result = """type Location;
fluent boolean robot_at(Location position);
fluent float [0.0, 100.0] battery_charge;
action move(Location l_from, Location l_to) {
   [ start ] (10 <= battery_charge);
   [ start ] (not (l_from == l_to));
   [ start ] robot_at(l_from);
   [ start ] (not robot_at(l_to));
   [ start ] robot_at(l_from) := false;
   [ start ] robot_at(l_to) := true;
   [ start ] battery_charge := (battery_charge - 10);
};
instance Location l1, l2;
[ start ] robot_at(l1) := true;
[ start ] robot_at(l2) := false;
[ start ] battery_charge := 100;
[ end ] robot_at(l2);
"""
        self.assertEqual(anml_problem, expected_result)

    def test_ad_hoc_1(self):
        when = UserType("when")
        fl = Fluent("4ction")
        obj = Object("predicate", when)
        act = InstantaneousAction("variable", fluent=when)
        fluent = act.parameter("fluent")
        act.add_effect(fl, True, Equals(fluent, obj))
        problem = Problem("ad_hoc")
        problem.add_fluent(fl)
        problem.add_action(act)
        problem.add_object(obj)
        problem.set_initial_value(fl, False)
        aw = ANMLWriter(problem)
        anml_problem = aw.get_problem()
        expected_result = """type when_;
fluent boolean f_4ction;
action variable_(when_ fluent_) {
   when [ start ] (fluent_ == predicate_)
   {[ start ] f_4ction := true;
   }
};
instance when_ predicate_;
[ start ] f_4ction := false;
"""
        self.assertEqual(anml_problem, expected_result)

    @skipIfEngineNotAvailable("tamer")
    def test_with_pytamer(self):
        import pytamer

        with tempfile.TemporaryDirectory() as tempdir:
            temp_file_name = os.path.join(tempdir, "test_file.anml")
            with OneshotPlanner(name="tamer") as tamer:
                for example in self.problems.values():
                    problem = example.problem
                    if tamer.supports(problem.kind):
                        aw = ANMLWriter(problem)
                        aw.write_problem(temp_file_name)
                        tamer_env = pytamer.tamer_env_new()
                        pytamer_problem = pytamer.tamer_parse_anml(
                            tamer_env, temp_file_name
                        )
                        tamer_actions = list(
                            pytamer.tamer_problem_get_actions(pytamer_problem)
                        )
                        self.assertEqual(len(tamer_actions), len(problem.actions))
                        for ta in tamer_actions:
                            up_act = problem.action(pytamer.tamer_action_get_name(ta))
                            ta_params = list(pytamer.tamer_action_get_parameters(ta))
                            self.assertEqual(len(up_act.parameters), len(ta_params))
                        tamer_fluents_and_constants = list(
                            pytamer.tamer_problem_get_fluents(pytamer_problem)
                        )
                        number_of_fluents = len(tamer_fluents_and_constants)
                        tamer_fluents_and_constants.extend(
                            list(pytamer.tamer_problem_get_constants(pytamer_problem))
                        )
                        self.assertEqual(
                            len(tamer_fluents_and_constants), len(problem.fluents)
                        )
                        for n, tf in enumerate(tamer_fluents_and_constants):
                            if n < number_of_fluents:  # the current element is a fluent
                                up_fluent = problem.fluent(
                                    pytamer.tamer_fluent_get_name(tf)
                                )
                                tf_sign = list(pytamer.tamer_fluent_get_parameters(tf))
                            else:  # The current element is a constant
                                up_fluent = problem.fluent(
                                    pytamer.tamer_constant_get_name(tf)
                                )
                                tf_sign = list(
                                    pytamer.tamer_constant_get_parameters(tf)
                                )
                            self.assertEqual(len(up_fluent.signature), len(tf_sign))
                        tamer_objects = list(
                            pytamer.tamer_problem_get_instances(pytamer_problem)
                        )
                        for to in tamer_objects:
                            problem.object(pytamer.tamer_instance_get_name(to))
                        self.assertEqual(len(tamer_objects), len(problem.all_objects))



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module defines the Environment class.
The Environment is a structure that contains multiple
singleton objects that are used throughout the system,
such as the ExpressionManager, TypeChecker, TypeManager.
"""


import sys
from typing import IO, Optional
import unified_planning


class Environment:
    """Represents the environment."""

    def __init__(self):
        import unified_planning.model
        import unified_planning.engines
        import unified_planning.model.walkers

        self._type_manager = unified_planning.model.TypeManager()
        self._factory = unified_planning.engines.Factory(self)
        self._tc = unified_planning.model.walkers.TypeChecker(self)
        self._expression_manager = unified_planning.model.ExpressionManager(self)
        self._free_vars_oracle = unified_planning.model.FreeVarsOracle()
        self._simplifier = unified_planning.model.walkers.Simplifier(self)
        self._free_vars_extractor = unified_planning.model.walkers.FreeVarsExtractor()
        self._credits_stream: Optional[IO[str]] = sys.stdout

    # The getstate and setstate method are needed in the Parallel engine. The
    #  Parallel engine creates a deep copy of the Environment instance in
    #  another process by pickling the enviroment fields.
    # Since the IO[str] class is not picklable, we need to remove it from the
    #  state and then add it as None in the new process
    def __getstate__(self):
        state = self.__dict__.copy()
        # Don't pickle _credits_stream
        del state["_credits_stream"]
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        # Add _credits_stream back since it doesn't exist in the pickle
        self._credits_stream = None

    @property
    def free_vars_oracle(self) -> "unified_planning.model.FreeVarsOracle":
        return self._free_vars_oracle

    @property
    def expression_manager(self) -> "unified_planning.model.ExpressionManager":
        return self._expression_manager

    @property
    def type_manager(self) -> "unified_planning.model.TypeManager":
        return self._type_manager

    @property
    def type_checker(self) -> "unified_planning.model.walkers.TypeChecker":
        """Get the Type Checker"""
        return self._tc

    @property
    def factory(self) -> "unified_planning.engines.Factory":
        return self._factory

    @property
    def simplifier(self) -> "unified_planning.model.walkers.Simplifier":
        return self._simplifier

    @property
    def free_vars_extractor(self) -> "unified_planning.model.walkers.FreeVarsExtractor":
        return self._free_vars_extractor

    @property
    def credits_stream(self) -> "Optional[IO[str]]":
        """Returns the stream where the engines credits are printed."""
        return self._credits_stream

    @credits_stream.setter
    def credits_stream(self, new_credits_stream: Optional[IO[str]]):
        """Sets the stream where the engines credits are printed."""
        self._credits_stream = new_credits_stream


GLOBAL_ENVIRONMENT: Optional[Environment] = None


def get_env(env: Environment = None) -> Environment:
    global GLOBAL_ENVIRONMENT
    if env is None:
        if GLOBAL_ENVIRONMENT is None:
            GLOBAL_ENVIRONMENT = Environment()
        return GLOBAL_ENVIRONMENT
    else:
        return env



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Provides the most used functions in a nicely wrapped API.
This module defines a global environment, so that most methods can be
called without the need to specify an environment or a ExpressionManager.
"""

import sys
import unified_planning as up
import unified_planning.model.types
from unified_planning.environment import get_env
from unified_planning.model import *
from unified_planning.engines import Engine, CompilationKind
from typing import IO, Iterable, List, Union, Dict, Tuple, Optional
from fractions import Fraction


def And(*args: Union[BoolExpression, Iterable[BoolExpression]]) -> FNode:
    return get_env().expression_manager.And(*args)


def Or(*args: Union[BoolExpression, Iterable[BoolExpression]]) -> FNode:
    return get_env().expression_manager.Or(*args)


def Not(expression: BoolExpression) -> FNode:
    return get_env().expression_manager.Not(expression)


def Implies(left: BoolExpression, right: BoolExpression) -> FNode:
    return get_env().expression_manager.Implies(left, right)


def Iff(left: BoolExpression, right: BoolExpression) -> FNode:
    return get_env().expression_manager.Iff(left, right)


def Exists(
    expression: BoolExpression, *vars: "unified_planning.model.Variable"
) -> FNode:
    return get_env().expression_manager.Exists(expression, *vars)


def Forall(
    expression: BoolExpression, *vars: "unified_planning.model.Variable"
) -> FNode:
    return get_env().expression_manager.Forall(expression, *vars)


def FluentExp(
    fluent: "unified_planning.model.Fluent", params: Tuple[Expression, ...] = tuple()
) -> FNode:
    return get_env().expression_manager.FluentExp(fluent, params)


def ParameterExp(param: "unified_planning.model.Parameter") -> FNode:
    return get_env().expression_manager.ParameterExp(param)


def VariableExp(var: "unified_planning.model.Variable") -> FNode:
    return get_env().expression_manager.VariableExp(var)


def ObjectExp(obj: "unified_planning.model.Object") -> FNode:
    return get_env().expression_manager.ObjectExp(obj)


def TRUE() -> FNode:
    return get_env().expression_manager.TRUE()


def FALSE() -> FNode:
    return get_env().expression_manager.FALSE()


def Bool(value: bool) -> FNode:
    return get_env().expression_manager.Bool(value)


def Int(value: int) -> FNode:
    return get_env().expression_manager.Int(value)


def Real(value: Fraction) -> FNode:
    return get_env().expression_manager.Real(value)


def Plus(*args: Union[Expression, Iterable[Expression]]) -> FNode:
    return get_env().expression_manager.Plus(*args)


def Minus(left: Expression, right: Expression) -> FNode:
    return get_env().expression_manager.Minus(left, right)


def Times(*args: Union[Expression, Iterable[Expression]]) -> FNode:
    return get_env().expression_manager.Times(*args)


def Div(left: Expression, right: Expression) -> FNode:
    return get_env().expression_manager.Div(left, right)


def LE(left: Expression, right: Expression) -> FNode:
    return get_env().expression_manager.LE(left, right)


def GE(left: Expression, right: Expression) -> FNode:
    return get_env().expression_manager.GE(left, right)


def LT(left: Expression, right: Expression) -> FNode:
    return get_env().expression_manager.LT(left, right)


def GT(left: Expression, right: Expression) -> FNode:
    return get_env().expression_manager.GT(left, right)


def Equals(left: Expression, right: Expression) -> FNode:
    return get_env().expression_manager.Equals(left, right)


def BoolType() -> unified_planning.model.types.Type:
    return get_env().type_manager.BoolType()


def IntType(
    lower_bound: int = None, upper_bound: int = None
) -> unified_planning.model.types.Type:
    return get_env().type_manager.IntType(lower_bound, upper_bound)


def RealType(
    lower_bound: Fraction = None, upper_bound: Fraction = None
) -> unified_planning.model.types.Type:
    return get_env().type_manager.RealType(lower_bound, upper_bound)


def UserType(
    name: str, father: Optional[Type] = None
) -> unified_planning.model.types.Type:
    return get_env().type_manager.UserType(name, father)


def OneshotPlanner(
    *,
    name: Optional[str] = None,
    names: Optional[List[str]] = None,
    params: Union[Dict[str, str], List[Dict[str, str]]] = None,
    problem_kind: ProblemKind = ProblemKind(),
    optimality_guarantee: Optional[Union["up.engines.OptimalityGuarantee", str]] = None
) -> Engine:
    """
    Returns a oneshot planner. There are three ways to call this method:
    - using 'name' (the name of a specific planner) and 'params' (planner dependent options).
      e.g. OneshotPlanner(name='tamer', params={'heuristic': 'hadd'})
    - using 'names' (list of specific planners name) and 'params' (list of
      planner dependent options) to get a Parallel engine.
      e.g. OneshotPlanner(names=['tamer', 'tamer'],
                          params=[{'heuristic': 'hadd'}, {'heuristic': 'hmax'}])
    - using 'problem_kind' and 'optimality_guarantee'.
          e.g. OneshotPlanner(problem_kind=problem.kind, optimality_guarantee=SOLVED_OPTIMALLY)
    """
    return get_env().factory.OneshotPlanner(
        name=name,
        names=names,
        params=params,
        problem_kind=problem_kind,
        optimality_guarantee=optimality_guarantee,
    )


def PlanValidator(
    *,
    name: Optional[str] = None,
    names: Optional[List[str]] = None,
    params: Union[Dict[str, str], List[Dict[str, str]]] = None,
    problem_kind: ProblemKind = ProblemKind(),
    plan_kind: Optional[Union["up.plans.PlanKind", str]] = None
) -> Engine:
    """
    Returns a plan validator. There are three ways to call this method:
    - using 'name' (the name of a specific plan validator) and 'params'
      (plan validator dependent options).
      e.g. PlanValidator(name='tamer', params={'opt': 'val'})
    - using 'names' (list of specific plan validators name) and 'params' (list of
      plan validators dependent options) to get a Parallel engine.
      e.g. PlanValidator(names=['tamer', 'tamer'],
                         params=[{'opt1': 'val1'}, {'opt2': 'val2'}])
    - using 'problem_kind' and 'plan_kind' parameters.
      e.g. PlanValidator(problem_kind=problem.kind, plan_kind=plan.kind)
    """
    return get_env().factory.PlanValidator(
        name=name,
        names=names,
        params=params,
        problem_kind=problem_kind,
        plan_kind=plan_kind,
    )


def Compiler(
    *,
    name: Optional[str] = None,
    params: Union[Dict[str, str], List[Dict[str, str]]] = None,
    problem_kind: ProblemKind = ProblemKind(),
    compilation_kind: Optional[Union["up.engines.CompilationKind", str]] = None
) -> "up.engines.engine.Engine":
    """
    Returns a Compiler. There are two ways to call this method:
    - using 'name' (the name of a specific grounder) and 'params'
        (grounder dependent options).
        e.g. Compiler(name='tamer', params={'opt': 'val'})
    - using 'problem_kind' and 'compilation_kind' parameters.
        e.g. Compiler(problem_kind=problem.kind, compilation_kind=GROUNDER)
    """
    return get_env().factory.Compiler(
        name=name,
        params=params,
        problem_kind=problem_kind,
        compilation_kind=compilation_kind,
    )


def Simulator(
    problem: "up.model.AbstractProblem",
    *,
    name: Optional[str] = None,
    params: Union[Dict[str, str], List[Dict[str, str]]] = None
) -> "up.engines.engine.Engine":
    """
    Returns a Simulator. There are two ways to call this method:
    - using 'problem_kind' through the problem field.
        e.g. Simulator(problem)
    - using 'name' (the name of a specific simulator) and eventually some 'params'
        (simulator dependent options).
        e.g. Simulator(problem, name='sequential_simulator')
    """
    return get_env().factory.Simulator(problem=problem, name=name, params=params)


def print_engines_info(stream: IO[str] = sys.stdout, full_credits: bool = False):
    get_env().factory.print_engines_info(stream, full_credits)


def set_credits_stream(stream: Optional[IO[str]]):
    get_env().credits_stream = stream



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import unified_planning as up
import unified_planning.plans as plans
from unified_planning.environment import Environment
from unified_planning.exceptions import UPUsageError
from typing import Callable, Optional, Tuple, List
from fractions import Fraction


class TimeTriggeredPlan(plans.plan.Plan):
    """Represents a time triggered plan."""

    def __init__(
        self,
        actions: List[Tuple[Fraction, "plans.plan.ActionInstance", Optional[Fraction]]],
        environment: Optional["Environment"] = None,
    ):
        """The first Fraction represents the absolute time in which the action
        Action starts, while the last Fraction represents the duration
        of the action to fullfill the problem goals.
        The Action can be an InstantaneousAction, this is represented with a duration set
        to None.
        """
        # if we have a specific env or we don't have any actions
        if environment is not None or not actions:
            plans.plan.Plan.__init__(
                self, plans.plan.PlanKind.TIME_TRIGGERED_PLAN, environment
            )
        # If we don't have a specific env and have at least 1 action, use the env of the first action
        else:
            assert len(actions) > 0
            plans.plan.Plan.__init__(
                self, plans.plan.PlanKind.TIME_TRIGGERED_PLAN, actions[0][1].action.env
            )
        for (
            _,
            ai,
            _,
        ) in actions:  # check that given env and the env in the actions is the same
            if ai.action.env != self._environment:
                raise UPUsageError(
                    "The environment given to the plan is not the same of the actions in the plan."
                )
        self._actions = actions

    def __repr__(self) -> str:
        return str(self._actions)

    def __eq__(self, oth: object) -> bool:
        if isinstance(oth, TimeTriggeredPlan) and len(self._actions) == len(
            oth._actions
        ):
            for (s, ai, d), (oth_s, oth_ai, oth_d) in zip(self._actions, oth._actions):
                if (
                    s != oth_s
                    or ai.action != oth_ai.action
                    or ai.actual_parameters != oth_ai.actual_parameters
                    or d != oth_d
                ):
                    return False
            return True
        else:
            return False

    def __hash__(self) -> int:
        count: int = 0
        for i, (s, ai, d) in enumerate(self._actions):
            count += (
                i + hash(ai.action) + hash(ai.actual_parameters) + hash(s) + hash(d)
            )
        return count

    def __contains__(self, item: object) -> bool:
        if isinstance(item, plans.plan.ActionInstance):
            return any(item.is_semantically_equivalent(a) for _, a, _ in self._actions)
        else:
            return False

    @property
    def timed_actions(
        self,
    ) -> List[Tuple[Fraction, "plans.plan.ActionInstance", Optional[Fraction]]]:
        """Returns the sequence of tuples (start, action_instance, duration) where:
        start is when the action starts;
        action_instance is the action applied;
        duration is the (optional) duration of the action."""
        return self._actions

    def replace_action_instances(
        self,
        replace_function: Callable[
            ["plans.plan.ActionInstance"], "plans.plan.ActionInstance"
        ],
    ) -> "plans.plan.Plan":
        new_ai = [(s, replace_function(ai), d) for s, ai, d in self._actions]
        new_env = self._environment
        if len(new_ai) > 0:
            _, ai, _ = new_ai[0]
            new_env = ai.action.env
        return TimeTriggeredPlan(new_ai, new_env)



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""This module defines an interface for a generic PDDL planner."""

import asyncio
from asyncio.subprocess import PIPE
import select
import subprocess
import sys
import tempfile
import os
import re
import time
import unified_planning as up
import unified_planning.engines as engines
import unified_planning.engines.mixins as mixins
from unified_planning.engines.results import (
    LogLevel,
    PlanGenerationResult,
    PlanGenerationResultStatus,
)
from unified_planning.io.pddl_writer import PDDLWriter
from unified_planning.exceptions import UPException
from asyncio.subprocess import PIPE
from typing import IO, Any, Callable, Optional, List, Tuple, Union, cast

# This module implements two different mechanisms to execute a PDDL planner in a
# subprocess, processing the output in real-time and imposing a timeout.
# The first one uses the "select" module to process the output and error streams
# while the subprocess is running. This method does not require multi-threading
# as it relies on the POSIX select to monitor multiple streams. Unfortunately,
# this method does not work in Windows because the select function only works
# with sockets. So, a second implementation uses asyncio futures to deal with
# the parallelism. This second method also has a limitation: it does not work in
# an environment that already uses asyncio (most notably in a google colab or in
# a python notebook).
#
# By default, on non-Windows OSs we use the first method and on Windows we
# always use the second. It is possible to use asyncio under unix by setting
# the environment variable UP_USE_ASYNCIO_PDDL_PLANNER to true.
USE_ASYNCIO_ON_UNIX = False
ENV_USE_ASYNCIO = os.environ.get("UP_USE_ASYNCIO_PDDL_PLANNER")
if ENV_USE_ASYNCIO is not None:
    USE_ASYNCIO_ON_UNIX = ENV_USE_ASYNCIO.lower() in ["true", "1"]


class PDDLPlanner(engines.engine.Engine, mixins.OneshotPlannerMixin):
    """
    This class is the interface of a generic PDDL planner
    that can be invocated through a subprocess call.
    """

    def __init__(self, needs_requirements=True):
        engines.engine.Engine.__init__(self)
        mixins.OneshotPlannerMixin.__init__(self)
        self._needs_requirements = needs_requirements

    def _get_cmd(
        self, domain_filename: str, problem_filename: str, plan_filename: str
    ) -> List[str]:
        """Takes in input two filenames where the problem's domain and problem are written, a
        filename where to write the plan and returns a list of command to run the engine on the
        problem and write the plan on the file called plan_filename."""
        raise NotImplementedError

    def _plan_from_file(
        self,
        problem: "up.model.Problem",
        plan_filename: str,
        get_item_named: Callable[
            [str],
            Union[
                "up.model.Type",
                "up.model.Action",
                "up.model.Fluent",
                "up.model.Object",
                "up.model.Parameter",
                "up.model.Variable",
            ],
        ],
    ) -> "up.plans.Plan":
        """Takes a problem, a filename and a map of renamings and returns the plan parsed from the file."""
        actions = []
        with open(plan_filename) as plan:
            for line in plan.readlines():
                if re.match(r"^\s*(;.*)?$", line):
                    continue
                res = re.match(r"^\s*\(\s*([\w?-]+)((\s+[\w?-]+)*)\s*\)\s*$", line)
                if res:
                    action = get_item_named(res.group(1))
                    assert isinstance(
                        action, up.model.Action
                    ), "Wrong plan or renaming."
                    parameters = []
                    for p in res.group(2).split():
                        obj = get_item_named(p)
                        assert isinstance(
                            obj, up.model.Object
                        ), "Wrong plan or renaming."
                        parameters.append(problem.env.expression_manager.ObjectExp(obj))
                    actions.append(up.plans.ActionInstance(action, tuple(parameters)))
                else:
                    raise UPException(
                        "Error parsing plan generated by " + self.__class__.__name__
                    )
        return up.plans.SequentialPlan(actions)

    def _solve(
        self,
        problem: "up.model.AbstractProblem",
        callback: Optional[
            Callable[["up.engines.results.PlanGenerationResult"], None]
        ] = None,
        timeout: Optional[float] = None,
        output_stream: Optional[IO[str]] = None,
    ) -> "up.engines.results.PlanGenerationResult":
        assert isinstance(problem, up.model.Problem)
        w = PDDLWriter(problem, self._needs_requirements)
        plan = None
        logs: List["up.engines.results.LogMessage"] = []
        with tempfile.TemporaryDirectory() as tempdir:
            domain_filename = os.path.join(tempdir, "domain.pddl")
            problem_filename = os.path.join(tempdir, "problem.pddl")
            plan_filename = os.path.join(tempdir, "plan.txt")
            w.write_domain(domain_filename)
            w.write_problem(problem_filename)
            cmd = self._get_cmd(domain_filename, problem_filename, plan_filename)

            if output_stream is None:
                # If we do not have an output stream to write to, we simply call
                # a subprocess and retrieve the final output and error with communicate
                process = subprocess.Popen(
                    cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE
                )
                timeout_occurred: bool = False
                proc_out: List[str] = []
                proc_err: List[str] = []
                try:
                    out_err_bytes = process.communicate(timeout=timeout)
                    proc_out, proc_err = [[x.decode()] for x in out_err_bytes]
                except subprocess.TimeoutExpired:
                    timeout_occurred = True
                retval = process.returncode
            else:
                if sys.platform == "win32":
                    # On windows we have to use asyncio (does not work inside notebooks)
                    try:
                        loop = asyncio.ProactorEventLoop()
                        exec_res = loop.run_until_complete(
                            run_command_asyncio(
                                cmd, output_stream=output_stream, timeout=timeout
                            )
                        )
                    finally:
                        loop.close()
                else:
                    # On non-windows OSs, we can choose between asyncio and posix
                    # select (see comment on USE_ASYNCIO_ON_UNIX variable for details)
                    if USE_ASYNCIO_ON_UNIX:
                        exec_res = asyncio.run(
                            run_command_asyncio(
                                cmd, output_stream=output_stream, timeout=timeout
                            )
                        )
                    else:
                        exec_res = run_command_posix_select(
                            cmd, output_stream=output_stream, timeout=timeout
                        )
                timeout_occurred, (proc_out, proc_err), retval = exec_res

            logs.append(up.engines.results.LogMessage(LogLevel.INFO, "".join(proc_out)))
            logs.append(
                up.engines.results.LogMessage(LogLevel.ERROR, "".join(proc_err))
            )
            if os.path.isfile(plan_filename):
                plan = self._plan_from_file(problem, plan_filename, w.get_item_named)
            if timeout_occurred and retval != 0:
                return PlanGenerationResult(
                    PlanGenerationResultStatus.TIMEOUT,
                    plan=plan,
                    log_messages=logs,
                    engine_name=self.name,
                )
        status: PlanGenerationResultStatus = self._result_status(problem, plan)
        return PlanGenerationResult(
            status, plan, log_messages=logs, engine_name=self.name
        )

    def _result_status(
        self, problem: "up.model.Problem", plan: Optional["up.plans.Plan"]
    ) -> "up.engines.results.PlanGenerationResultStatus":
        """Takes a problem and a plan and returns the status that represents this plan.
        The possible status with their interpretation can be found in the up.engines.results file."""
        raise NotImplementedError


async def run_command_asyncio(
    cmd: List[str], output_stream: IO[str], timeout: Optional[float] = None
) -> Tuple[bool, Tuple[List[str], List[str]], int]:
    """
    Executed the specified command line using asyncio primitives, imposing the specified timeout and printing online the output on output_stream.
    The function returns a boolean flag telling if a timeout occurred, a pair of string lists containing the captured standard output and standard error and the return code of the command as an integer
    """
    start = time.time()
    process = await asyncio.create_subprocess_exec(*cmd, stdout=PIPE, stderr=PIPE)

    timeout_occurred = False
    process_output: Tuple[List[str], List[str]] = ([], [])  # stdout, stderr
    while True:
        lines = [b"", b""]
        oks = [True, True]
        for idx, stream in enumerate([process.stdout, process.stderr]):
            assert stream is not None
            try:
                lines[idx] = await asyncio.wait_for(stream.readline(), 0.5)
            except asyncio.TimeoutError:
                oks[idx] = False

        if all(oks) and (not lines[0] and not lines[1]):  # EOF
            break
        else:
            for idx in range(2):
                output_string = lines[idx].decode().replace("\r\n", "\n")
                output_stream.write(output_string)
                process_output[idx].append(output_string)
        if timeout is not None and time.time() - start >= timeout:
            try:
                process.kill()
            except OSError:
                pass  # This can happen if the process is already terminated
            timeout_occurred = True
            break

    await process.wait()  # Wait for the child process to exit
    return timeout_occurred, process_output, cast(int, process.returncode)


def run_command_posix_select(
    cmd: List[str], output_stream: IO[str], timeout: Optional[float] = None
) -> Tuple[bool, Tuple[List[str], List[str]], int]:
    """
    Executed the specified command line using posix select, imposing the specified timeout and printing online the output on output_stream.
    The function returns a boolean flag telling if a timeout occurred, a pair of string lists containing the captured standard output and standard error and the return code of the command as an integer

    WARNING: this does not work under Windows because the select function only support sockets and not pipes
    WARNING: The resolution of the timeout parameter is ~ 1 second if output_stream is specified
    """
    proc_out: List[str] = []
    proc_err: List[str] = []
    proc_out_buff: List[str] = []
    proc_err_buff: List[str] = []

    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    timeout_occurred: bool = False
    start_time = time.time()
    last_red_out, last_red_err = 0, 0  # Variables needed for the correct loop exit
    readable_streams: List[Any] = []
    # Exit loop condition: Both stream have nothing left to read or the planner is out of time
    while not timeout_occurred and (
        len(readable_streams) != 2 or last_red_out != 0 or last_red_err != 0
    ):
        readable_streams, _, _ = select.select(
            [process.stdout, process.stderr], [], [], 1.0
        )  # 1.0 is the timeout resolution
        if (
            timeout is not None and time.time() - start_time >= timeout
        ):  # Check if the planner is out of time.
            try:
                process.kill()
            except OSError:
                pass  # This can happen if the process is already terminated
            timeout_occurred = True
        for readable_stream in readable_streams:
            out_in_bytes = readable_stream.read()
            out_str = out_in_bytes.decode().replace("\r\n", "\n")
            output_stream.write(out_str)
            if readable_stream == process.stdout:
                last_red_out = len(out_in_bytes)
                buff = proc_out_buff
                lst = proc_out
            else:
                last_red_err = len(out_in_bytes)
                buff = proc_err_buff
                lst = proc_err
            buff.append(out_str)
            if "\n" in out_str:
                lines = "".join(buff).split("\n")
                for x in lines[:-1]:
                    lst.append(x + "\n")

                buff.clear()
                if lines[-1]:
                    buff.append(lines[-1])
        process.wait()
        lastout = "".join(proc_out_buff)
        if lastout:
            proc_out.append(lastout + "\n")
        lasterr = "".join(proc_err_buff)
        if lasterr:
            proc_err.append(lasterr + "\n")
    return timeout_occurred, (proc_out, proc_err), cast(int, process.returncode)



# Copyright 2022 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module defines the Method class.
A Method has a name, a list of Parameters, a list of conditions
and a list of subtasks.
"""
from collections import OrderedDict
from typing import List, Union, Optional

import unified_planning as up
from unified_planning.environment import get_env, Environment
from unified_planning.exceptions import UPUnboundedVariablesError, UPValueError
from unified_planning.model.parameter import Parameter
from unified_planning.model.action import Action
from unified_planning.model.htn.task import Task, Subtask
from unified_planning.model.timing import Timing, Timepoint
from unified_planning.model.expression import Expression


class ParameterizedTask:
    """A task instantiated with some parameters."""

    def __init__(self, task: Task, *params: Parameter):
        self._task = task
        self._params: List[Parameter] = list(params)
        assert len(self._task.parameters) == len(self._params)
        # TODO #153: check that the type of each parameter is compatible with the task' signature

    def __repr__(self):
        return str(self._task.name) + "(" + ", ".join(map(str, self.parameters)) + ")"

    def __eq__(self, other):
        return (
            isinstance(other, ParameterizedTask)
            and self._task == other._task
            and self._params == other._params
        )

    def __hash__(self):
        return hash(self._task) + sum(map(hash, self._params))

    @property
    def task(self) -> Task:
        return self._task

    @property
    def parameters(self) -> List[up.model.parameter.Parameter]:
        return self._params


class Method:
    """This is the method interface."""

    def __init__(
        self,
        _name: str,
        _parameters: "Union[OrderedDict[str, up.model.types.Type], List[Parameter]]" = None,
        _env: Environment = None,
        **kwargs: "up.model.types.Type",
    ):
        self._env = get_env(_env)
        self._task: Optional[ParameterizedTask] = None
        self._name = _name
        self._parameters: "OrderedDict[str, Parameter]" = OrderedDict()
        self._preconditions: List[up.model.fnode.FNode] = []
        self._constraints: List[up.model.fnode.FNode] = []
        self._subtasks: List[Subtask] = []
        if _parameters is None:
            for n, t in kwargs.items():
                self._parameters[n] = Parameter(n, t, self._env)
        elif isinstance(_parameters, List):
            assert len(kwargs) == 0
            for p in _parameters:
                self._parameters[p.name] = p
        else:
            assert isinstance(_parameters, OrderedDict)
            assert len(kwargs) == 0
            for n, t in _parameters.items():
                self._parameters[n] = Parameter(n, t, self._env)

    def __repr__(self) -> str:
        s = []
        s.append(f"method {self.name}")
        first = True
        for p in self.parameters:
            if first:
                s.append("(")
                first = False
            else:
                s.append(", ")
            s.append(str(p))
        if not first:
            s.append(")")
        s.append(" {\n")
        s.append(f"  task = {self._task}\n")
        s.append("  preconditions = [\n")
        for c in self.preconditions:
            s.append(f"    {str(c)}\n")
        s.append("  ]\n")
        s.append("  constraints = [\n")
        for c in self.constraints:
            s.append(f"    {str(c)}\n")
        s.append("  ]\n")
        s.append("  subtasks = [\n")
        for st in self.subtasks:
            s.append(f"      {str(st)}\n")
        s.append("  ]\n")
        s.append("}")
        return "".join(s)

    def __eq__(self, oth: object) -> bool:
        if not isinstance(oth, Method):
            return False
        return (
            self._env == oth._env
            and self._name == oth._name
            and self._parameters == oth._parameters
            and self._task == oth._task
            and set(self._preconditions) == set(oth._preconditions)
            and set(self._subtasks) == set(oth._subtasks)
            and set(self._constraints) == set(oth._constraints)
        )

    def __hash__(self) -> int:
        res = hash(self._name)
        res += hash(self._task)
        res += sum(map(hash, self.parameters))
        res += sum(map(hash, self._preconditions))
        res += sum(map(hash, self._constraints))
        res += sum(map(hash, self.subtasks))
        return res

    @property
    def name(self) -> str:
        """Returns the action name."""
        return self._name

    @property
    def achieved_task(self) -> ParameterizedTask:
        """Returns the task that this method achieves."""
        assert (
            self._task is not None
        ), "The achieved task was previously set (see the set_task method)."
        return self._task

    def set_task(self, task: Union[Task, ParameterizedTask], *arguments: Parameter):
        """Defines the task that is method achieves.

        It expects a Task and its arguments, either bundle in a `ParameterizedTask` instance of
        passed separetly.
        It is assumed that each parameter of the achieved task is a parameter of the method.

        # Examples
        >>> from unified_planning.shortcuts import *
        >>> from unified_planning.model.htn import *
        >>> Location = UserType("Location")
        >>> go = Task("go", target=Location)
        >>> m1 = Method("m-go1", target=Location)
        >>> task_achieved = ParameterizedTask(go, m1.parameter("target"))
        >>> m1.set_task(task_achieved)
        >>> m2 = Method("m-go2", source=Location, target=Location)
        >>> m2.set_task(go, m2.parameter("target"))
        >>> m3 = Method("m-go3", source=Location, target=Location)
        >>> m3.set_task(go) # Infer the parameters of the `go` task from the parameters of m3 with the same name
        """
        assert self._task is None, f"Method {self.name} was already assigned a task"
        if isinstance(task, ParameterizedTask):
            assert (
                len(arguments) == 0
            ), "Unexpected arguments passed along a ParameterizedTask"
            assert all(
                p in self.parameters for p in task.parameters
            ), "A parameter of the task does not appear as a parameter of the method."
            self._task = task
        elif isinstance(task, Task) and len(arguments) == 0:
            for task_param in task.parameters:
                assert (
                    task_param.name in self._parameters
                ), f"Missing task parameter '{task_param.name}' in method {self._name}. Please pass all parameters explicitly."
            self._task = ParameterizedTask(task, *task.parameters)
        else:
            assert all(
                p in self.parameters for p in arguments
            ), "An argument passed to the task does not appear as a parameter of the method."
            self._task = ParameterizedTask(task, *arguments)

    @property
    def parameters(self) -> List[Parameter]:
        """Returns the list of the method's parameters."""
        return list(self._parameters.values())

    def parameter(self, name: str) -> Parameter:
        """Returns the parameter of the action with the given name."""
        for param in self.parameters:
            if param.name == name:
                return param
        raise UPValueError(f"Unknown parameter name: {name}")

    @property
    def preconditions(self) -> List["up.model.fnode.FNode"]:
        """Returns the list of the method's preconditions."""
        return self._preconditions

    def add_precondition(self, precondition: Expression):
        """Adds the given method precondition."""
        (precondition_exp,) = self._env.expression_manager.auto_promote(precondition)
        assert self._env.type_checker.get_type(precondition_exp).is_bool_type()
        if precondition_exp == self._env.expression_manager.TRUE():
            return
        free_vars = self._env.free_vars_oracle.get_free_variables(precondition_exp)
        if len(free_vars) != 0:
            raise UPUnboundedVariablesError(
                f"The precondition {str(precondition_exp)} has unbounded variables:\n{str(free_vars)}"
            )
        if precondition_exp not in self._preconditions:
            self._preconditions.append(precondition_exp)

    @property
    def constraints(self) -> List["up.model.fnode.FNode"]:
        """Returns the list of the method's constraints."""
        return self._constraints

    def add_constraint(self, constraint: Expression):
        """Adds the given constraint to the method."""
        (constraint_exp,) = self._env.expression_manager.auto_promote(constraint)
        assert self._env.type_checker.get_type(constraint_exp).is_bool_type()
        if constraint_exp == self._env.expression_manager.TRUE():
            return
        free_vars = self._env.free_vars_oracle.get_free_variables(constraint_exp)
        if len(free_vars) != 0:
            raise UPUnboundedVariablesError(
                f"The constraint {str(constraint_exp)} has unbounded variables:\n{str(free_vars)}"
            )
        if constraint_exp not in self._constraints:
            self._constraints.append(constraint_exp)

    @property
    def subtasks(self) -> List["Subtask"]:
        """Returns the list of the method's subtasks."""
        return self._subtasks

    def add_subtask(
        self,
        subtask: Union[Subtask, Action, Task],
        *args: Expression,
        ident: Optional[str] = None,
    ) -> Subtask:
        """Adds a subtask to this method, with no particular ordering relative to the existing ones."""
        if not isinstance(subtask, Subtask):
            parameters = self._env.expression_manager.auto_promote(*args)
            subtask = Subtask(subtask, *parameters, ident=ident)
        else:
            assert len(args) == 0
        assert isinstance(subtask, Subtask)
        assert all([subtask.identifier != prev.identifier for prev in self.subtasks])
        self._subtasks.append(subtask)
        return subtask

    def set_ordered(self, *subtasks: Subtask):
        """Imposes a sequential order between the given subtasks."""
        if len(subtasks) < 2:
            return
        prev = subtasks[0]
        for i in range(1, len(subtasks)):
            next = subtasks[i]
            self.set_strictly_before(prev, next)
            prev = next

    def set_strictly_before(
        self,
        lhs: Union[Subtask, Timepoint, Timing],
        rhs: Union[Subtask, Timepoint, Timing],
    ):
        if isinstance(lhs, Subtask):
            lhs = lhs.end
        if isinstance(lhs, Timepoint):
            lhs = Timing(timepoint=lhs, delay=0)
        assert isinstance(lhs, Timing)
        if isinstance(rhs, Subtask):
            rhs = rhs.start
        if isinstance(rhs, Timepoint):
            rhs = Timing(timepoint=rhs, delay=0)
        assert isinstance(rhs, Timing)
        self.add_constraint(self._env.expression_manager.LT(lhs, rhs))



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import networkx as nx
import unified_planning as up
import unified_planning.plans as plans
from unified_planning.environment import Environment
from unified_planning.exceptions import UPUsageError
from unified_planning.plans.plan import ActionInstance
from unified_planning.plans.sequential_plan import SequentialPlan
from typing import Callable, Dict, Iterator, List, Optional


class PartialOrderPlan(plans.plan.Plan):
    """Represents a partial order plan. Actions are represent as an adjacency list graph."""

    def __init__(
        self,
        adjacency_list: Dict[
            "plans.plan.ActionInstance", List["plans.plan.ActionInstance"]
        ],
        environment: Optional["Environment"] = None,
        _graph: Optional[nx.DiGraph] = None,
    ):
        """Constructs the PartialOrderPlan using the adjacency list representation.

        :param adjacency_list: The Dictionary representing the adjacency list for this PartialOrderPlan.
        :param env: The environment in which the ActionInstances in the adjacency_list are created.
        :param _graph: The graph that is semnatically equivalent to the adjacency_list.
            NOTE: This parameter is for internal use only and it's maintainance is not guaranteed by any means.
        :return: The created PartialOrderPlan.
        """
        # if we have a specific env or we don't have any actions
        if environment is not None or not adjacency_list:
            plans.plan.Plan.__init__(
                self, plans.plan.PlanKind.PARTIAL_ORDER_PLAN, environment
            )
        # If we don't have a specific env, use the env of the first action
        else:
            assert len(adjacency_list) > 0
            for ai in adjacency_list.keys():
                plans.plan.Plan.__init__(
                    self, plans.plan.PlanKind.PARTIAL_ORDER_PLAN, ai.action.env
                )
                break
        if _graph is not None:
            # sanity checks
            assert len(adjacency_list) == 0
            assert all(isinstance(n, ActionInstance) for n in _graph.nodes)
            assert all(
                isinstance(f, ActionInstance) and isinstance(t, ActionInstance)
                for f, t in _graph.edges
            )
            self._graph = _graph
        else:
            for (
                ai_k,
                ai_v_list,
            ) in (
                adjacency_list.items()
            ):  # check that given env and the env in the actions is the same
                if ai_k.action.env != self._environment:
                    raise UPUsageError(
                        "The environment given to the plan is not the same of the actions in the plan."
                    )
                for ai in ai_v_list:
                    if ai.action.env != self._environment:
                        raise UPUsageError(
                            "The environment given to the plan is not the same of the actions in the plan."
                        )
            self._graph = nx.convert.from_dict_of_lists(
                adjacency_list, create_using=nx.DiGraph
            )

    def __repr__(self) -> str:
        return str(self._graph)

    def __eq__(self, oth: object) -> bool:
        if isinstance(oth, PartialOrderPlan):
            return nx.is_isomorphic(
                self._graph,
                oth._graph,
                node_match=_semantically_equivalent_action_instances,
            )
        else:
            return False

    def __hash__(self) -> int:
        return hash(nx.weisfeiler_lehman_graph_hash(self._graph))

    def __contains__(self, item: object) -> bool:
        if isinstance(item, ActionInstance):
            return any(item.is_semantically_equivalent(a) for a in self._graph.nodes)
        else:
            return False

    @property
    def get_adjacency_list(
        self,
    ) -> Dict["plans.plan.ActionInstance", List["plans.plan.ActionInstance"]]:
        """Returns the graph of action instances as an adjacency list."""
        return nx.convert.to_dict_of_lists(self._graph)

    def replace_action_instances(
        self,
        replace_function: Callable[
            ["plans.plan.ActionInstance"], "plans.plan.ActionInstance"
        ],
    ) -> "plans.plan.Plan":
        new_adj_list: Dict[
            "plans.plan.ActionInstance", List["plans.plan.ActionInstance"]
        ] = {}
        # Populate the new adjacency list with the replaced action instances
        for node in self._graph.nodes:
            new_adj_list[replace_function(node)] = [
                replace_function(successor) for successor in self._graph.neighbors(node)
            ]
        new_env = self._environment
        for ai in new_adj_list.keys():
            new_env = ai.action.env
            break
        return PartialOrderPlan(new_adj_list, new_env)

    def to_sequential_plan(self) -> SequentialPlan:
        """Returns one between all possible SequentialPlans that respects the ordering constaints given by this PartialOrderPlan."""
        return SequentialPlan(list(nx.topological_sort(self._graph)), self._environment)

    def all_sequential_plans(self) -> Iterator[SequentialPlan]:
        """Returns all possible SequentialPlans that respects the ordering constaints given by this PartialOrderPlan."""
        for sorted_plan in nx.all_topological_sorts(self._graph):
            yield SequentialPlan(list(sorted_plan), self._environment)

    def get_neighbors(
        self, action_instance: ActionInstance
    ) -> Iterator[ActionInstance]:
        try:
            retval = self._graph.neighbors(action_instance)
        except nx.NetworkXError:
            raise UPUsageError(
                f"The action instance {str(action_instance)} does not belong to this Partial Order Plan. \n Note that 2 Action Instances are equals if and only if they are the exact same object."
            )
        return retval


def _semantically_equivalent_action_instances(
    action_instance_1: ActionInstance, action_instance_2: ActionInstance
) -> bool:
    return action_instance_1.is_semantically_equivalent(action_instance_2)
